"""Tests for the sop_class module."""

import pytest

from pynetdicom import __version__
from pydicom._uid_dict import UID_dictionary

from pynetdicom import sop_class
from pynetdicom.sop_class import (
    uid_to_sop_class,
    uid_to_service_class,
    SOPClass,
    _SERVICE_CLASSES,
    _APPLICATION_EVENT_CLASSES,
    ProceduralEventLogging,
    _BASIC_WORKLIST_CLASSES,
    ModalityWorklistInformationFind,
    _COLOR_PALETTE_CLASSES,
    ColorPaletteInformationModelMove,
    _DEFINED_PROCEDURE_CLASSES,
    DefinedProcedureProtocolInformationModelFind,
    _DISPLAY_SYSTEM_CLASSES,
    DisplaySystem,
    _HANGING_PROTOCOL_CLASSES,
    HangingProtocolInformationModelGet,
    _IMPLANT_TEMPLATE_CLASSES,
    ImplantTemplateGroupInformationModelFind,
    _INSTANCE_AVAILABILITY_CLASSES,
    InstanceAvailabilityNotification,
    _INVENTORY_CLASSES,
    InventoryFind,
    _MEDIA_CREATION_CLASSES,
    MediaCreationManagement,
    _MEDIA_STORAGE_CLASSES,
    MediaStorageDirectoryStorage,
    _NON_PATIENT_OBJECT_CLASSES,
    HangingProtocolStorage,
    _PRINT_MANAGEMENT_CLASSES,
    PrintJob,
    _PROCEDURE_STEP_CLASSES,
    ModalityPerformedProcedureStep,
    _PROTOCOL_APPROVAL_CLASSES,
    ProtocolApprovalInformationModelFind,
    ProtocolApprovalInformationModelMove,
    ProtocolApprovalInformationModelGet,
    _QR_CLASSES,
    StudyRootQueryRetrieveInformationModelFind,
    _RELEVANT_PATIENT_QUERY_CLASSES,
    GeneralRelevantPatientInformationQuery,
    _RT_MACHINE_VERIFICATION_CLASSES,
    RTConventionalMachineVerification,
    _STORAGE_CLASSES,
    CTImageStorage,
    _STORAGE_COMMITMENT_CLASSES,
    StorageCommitmentPushModel,
    _STORAGE_MANAGEMENT_CLASSES,
    InventoryCreation,
    _SUBSTANCE_ADMINISTRATION_CLASSES,
    ProductCharacteristicsQuery,
    _UNIFIED_PROCEDURE_STEP_CLASSES,
    UnifiedProcedureStepPull,
    _VERIFICATION_CLASSES,
    Verification,
    DisplaySystemInstance,
    PrinterConfigurationRetrievalInstance,
    PrinterInstance,
    ProceduralEventLoggingInstance,
    StorageCommitmentPushModelInstance,
    StorageManagementInstance,
    SubstanceAdministrationLoggingInstance,
    UPSFilteredGlobalSubscriptionInstance,
    UPSGlobalSubscriptionInstance,
    register_uid,
)
from pynetdicom.service_class import (
    ServiceClass,
    BasicWorklistManagementServiceClass,
    ColorPaletteQueryRetrieveServiceClass,
    DefinedProcedureProtocolQueryRetrieveServiceClass,
    HangingProtocolQueryRetrieveServiceClass,
    ImplantTemplateQueryRetrieveServiceClass,
    InventoryQueryRetrieveServiceClass,
    NonPatientObjectStorageServiceClass,
    ProtocolApprovalQueryRetrieveServiceClass,
    QueryRetrieveServiceClass,
    RelevantPatientInformationQueryServiceClass,
    StorageServiceClass,
    SubstanceAdministrationQueryServiceClass,
    VerificationServiceClass,
)
from pynetdicom.service_class_n import (
    ApplicationEventLoggingServiceClass,
    DisplaySystemManagementServiceClass,
    InstanceAvailabilityNotificationServiceClass,
    MediaCreationManagementServiceClass,
    PrintManagementServiceClass,
    ProcedureStepServiceClass,
    RTMachineVerificationServiceClass,
    StorageCommitmentServiceClass,
    StorageManagementServiceClass,
    UnifiedProcedureStepServiceClass,
)


PYDICOM_VERSION = __version__.split(".")[:2]


def test_all_sop_classes():
    """Test that all the SOP Class UIDs are correct."""
    for uid in _APPLICATION_EVENT_CLASSES.values():
        assert uid in UID_dictionary
    for uid in _BASIC_WORKLIST_CLASSES.values():
        assert uid in UID_dictionary
    for uid in _COLOR_PALETTE_CLASSES.values():
        assert uid in UID_dictionary
    for uid in _DEFINED_PROCEDURE_CLASSES.values():
        assert uid in UID_dictionary
    for uid in _DISPLAY_SYSTEM_CLASSES.values():
        assert uid in UID_dictionary
    for uid in _HANGING_PROTOCOL_CLASSES.values():
        assert uid in UID_dictionary
    for uid in _IMPLANT_TEMPLATE_CLASSES.values():
        assert uid in UID_dictionary
    for uid in _INSTANCE_AVAILABILITY_CLASSES.values():
        assert uid in UID_dictionary
    for uid in _INVENTORY_CLASSES.values():
        assert uid in UID_dictionary
    for uid in _MEDIA_CREATION_CLASSES.values():
        assert uid in UID_dictionary
    for uid in _MEDIA_STORAGE_CLASSES.values():
        assert uid in UID_dictionary
    for uid in _NON_PATIENT_OBJECT_CLASSES.values():
        assert uid in UID_dictionary
    for uid in _PRINT_MANAGEMENT_CLASSES.values():
        assert uid in UID_dictionary
    for uid in _PROCEDURE_STEP_CLASSES.values():
        assert uid in UID_dictionary
    for uid in _PROTOCOL_APPROVAL_CLASSES.values():
        assert uid in UID_dictionary
    for uid in _QR_CLASSES.values():
        assert uid in UID_dictionary
    for uid in _RELEVANT_PATIENT_QUERY_CLASSES.values():
        assert uid in UID_dictionary
    for uid in _RT_MACHINE_VERIFICATION_CLASSES.values():
        assert uid in UID_dictionary
    for uid in _STORAGE_CLASSES.values():
        if uid in (
            "1.2.840.10008.5.1.4.1.1.6.3",
            "1.2.840.10008.5.1.4.1.1.9.1.4",
            "1.2.840.10008.5.1.4.1.1.77.1.8",
            "1.2.840.10008.5.1.4.1.1.77.1.9",
        ):
            continue

        assert uid in UID_dictionary
    for uid in _STORAGE_COMMITMENT_CLASSES.values():
        assert uid in UID_dictionary
    for uid in _STORAGE_MANAGEMENT_CLASSES.values():
        assert uid in UID_dictionary
    for uid in _SUBSTANCE_ADMINISTRATION_CLASSES.values():
        assert uid in UID_dictionary
    for uid in _UNIFIED_PROCEDURE_STEP_CLASSES.values():
        assert uid in UID_dictionary
    for uid in _VERIFICATION_CLASSES.values():
        assert uid in UID_dictionary


def test_all_service_classes():
    """Test that all the Service Class UIDs are correct."""
    for uid in _SERVICE_CLASSES:
        assert uid in UID_dictionary


def test_all_sop_instances():
    """Test the well-known SOP Instances are correct."""
    assert DisplaySystemInstance in UID_dictionary
    assert PrinterConfigurationRetrievalInstance in UID_dictionary
    assert PrinterInstance in UID_dictionary
    assert ProceduralEventLoggingInstance in UID_dictionary
    assert StorageCommitmentPushModelInstance in UID_dictionary
    assert StorageManagementInstance in UID_dictionary
    assert SubstanceAdministrationLoggingInstance in UID_dictionary
    assert UPSFilteredGlobalSubscriptionInstance in UID_dictionary
    assert UPSGlobalSubscriptionInstance in UID_dictionary


class TestUIDtoSOPlass:
    """Tests for uid_to_sop_class"""

    def test_missing_sop_class(self):
        """Test SOP Class if UID not found."""
        sop = uid_to_sop_class("1.2.3.4")
        assert sop == "1.2.3.4"
        assert sop.service_class == ServiceClass

    def test_verification_uid(self):
        """Test normal function"""
        assert uid_to_sop_class("1.2.840.10008.1.1") == Verification

    def test_existing(self):
        """Test that the existing class is returned."""
        original = Verification
        sop = uid_to_sop_class("1.2.840.10008.1.1")
        assert id(sop) == id(original)


class TestUIDToServiceClass:
    """Tests for sop_class.uid_to_service_class."""

    def test_service_class_uid(self):
        uid = "1.2.840.10008.4.2"
        assert uid_to_service_class(uid) == StorageServiceClass

    def test_app_logging_uids(self):
        """Test that the Application Event SOP Class UIDs work correctly."""
        for uid in _APPLICATION_EVENT_CLASSES.values():
            assert uid_to_service_class(uid) == ApplicationEventLoggingServiceClass

    def test_basic_worklist_uids(self):
        """Test that the Basic Worklist SOP Class UIDs work correctly."""
        for uid in _BASIC_WORKLIST_CLASSES.values():
            assert uid_to_service_class(uid) == BasicWorklistManagementServiceClass

    def test_color_palette_uids(self):
        """Test that the Color Paletter SOP Class UIDs work correctly."""
        for uid in _COLOR_PALETTE_CLASSES.values():
            assert uid_to_service_class(uid) == ColorPaletteQueryRetrieveServiceClass

    def test_defined_procedure_uids(self):
        """Test that the Defined Procedure SOP Class UIDs work correctly."""
        for uid in _DEFINED_PROCEDURE_CLASSES.values():
            assert (
                uid_to_service_class(uid)
                == DefinedProcedureProtocolQueryRetrieveServiceClass
            )

    def test_display_system_uids(self):
        """Test that Display System SOP Class UIDs work correctly."""
        for uid in _DISPLAY_SYSTEM_CLASSES.values():
            assert uid_to_service_class(uid) == DisplaySystemManagementServiceClass

    def test_hanging_protocol_uids(self):
        """Test that the Hanging Protocol SOP Class UIDs work correctly."""
        for uid in _HANGING_PROTOCOL_CLASSES.values():
            assert uid_to_service_class(uid) == HangingProtocolQueryRetrieveServiceClass

    def test_implant_template_uids(self):
        """Test that the Implant Template SOP Class UIDs work correctly."""
        for uid in _IMPLANT_TEMPLATE_CLASSES.values():
            assert uid_to_service_class(uid) == ImplantTemplateQueryRetrieveServiceClass

    def test_instance_uids(self):
        """Test that the Instance Availability SOP Class UIDs work correctly."""
        for uid in _INSTANCE_AVAILABILITY_CLASSES.values():
            assert (
                uid_to_service_class(uid)
                == InstanceAvailabilityNotificationServiceClass
            )

    def test_inventory_uids(self):
        """Test that the Inventory QR SOP Class UIDs work correctly."""
        for uid in _INVENTORY_CLASSES.values():
            assert uid_to_service_class(uid) == InventoryQueryRetrieveServiceClass

    def test_media_creation_uids(self):
        """Test that the Media Creation SOP Class UIDs work correctly."""
        for uid in _MEDIA_CREATION_CLASSES.values():
            assert uid_to_service_class(uid) == MediaCreationManagementServiceClass

    def test_media_storage_uids(self):
        """Test that the Media Storage SOP Class UIDs work correctly."""
        for uid in _MEDIA_STORAGE_CLASSES.values():
            assert uid_to_service_class(uid) == ServiceClass

    def test_non_patient_uids(self):
        """Test that the Non-Patient Object SOP Class UIDs work correctly."""
        for uid in _NON_PATIENT_OBJECT_CLASSES.values():
            assert uid_to_service_class(uid) == NonPatientObjectStorageServiceClass

    def test_print_uids(self):
        """Test that the Print SOP Class UIDs work correctly."""
        for uid in _PRINT_MANAGEMENT_CLASSES.values():
            assert uid_to_service_class(uid) == PrintManagementServiceClass

    def test_procedure_step_uids(self):
        """Test that the Procedure SOP Class UIDs work correctly."""
        for uid in _PROCEDURE_STEP_CLASSES.values():
            assert uid_to_service_class(uid) == ProcedureStepServiceClass

    def test_protocol_approval_uids(self):
        """Test that Protocol Approval SOP Class UIDs work correctly."""
        for uid in _PROTOCOL_APPROVAL_CLASSES.values():
            assert (
                uid_to_service_class(uid) == ProtocolApprovalQueryRetrieveServiceClass
            )

    def test_qr_uids(self):
        """Test that the QR SOP Class UIDs work correctly."""
        for uid in _QR_CLASSES.values():
            assert uid_to_service_class(uid) == QueryRetrieveServiceClass

    def test_relevant_patient_uids(self):
        """Test that the Relevant Patient SOP Class UIDs work correctly."""
        for uid in _RELEVANT_PATIENT_QUERY_CLASSES.values():
            assert (
                uid_to_service_class(uid) == RelevantPatientInformationQueryServiceClass
            )

    def test_rt_machine_uids(self):
        """Test that the RT Verification SOP Class UIDs work correctly."""
        for uid in _RT_MACHINE_VERIFICATION_CLASSES.values():
            assert uid_to_service_class(uid) == RTMachineVerificationServiceClass

    def test_storage_uids(self):
        """Test that the Storage SOP Class UIDs work correctly."""
        for uid in _STORAGE_CLASSES.values():
            assert uid_to_service_class(uid) == StorageServiceClass

    def test_storage_commitment_uids(self):
        """Test that the Storage Commitment SOP Class UIDs work correctly."""
        for uid in _STORAGE_COMMITMENT_CLASSES.values():
            assert uid_to_service_class(uid) == StorageCommitmentServiceClass

    def test_storage_management_uids(self):
        """Test that the Storage Management SOP Class UIDs work correctly."""
        for uid in _STORAGE_MANAGEMENT_CLASSES.values():
            assert uid_to_service_class(uid) == StorageManagementServiceClass

    def test_substance_admin_uids(self):
        """Test that the Substance Administration SOP Class UIDs work correctly."""
        for uid in _SUBSTANCE_ADMINISTRATION_CLASSES.values():
            assert uid_to_service_class(uid) == SubstanceAdministrationQueryServiceClass

    def test_ups_uids(self):
        """Test that the UPS SOP Class UIDs work correctly."""
        for uid in _UNIFIED_PROCEDURE_STEP_CLASSES.values():
            assert uid_to_service_class(uid) == UnifiedProcedureStepServiceClass

    def test_verification_uids(self):
        """Test that the Verification SOP Class UIDs work correctly."""
        for uid in _VERIFICATION_CLASSES.values():
            assert uid_to_service_class(uid) == VerificationServiceClass

    def test_unknown_uid(self):
        """Test that an unknown UID returns default service class."""
        assert uid_to_service_class("1.2.3") == ServiceClass


class TestSOPClass:
    """Tests for sop_class.SOPClass."""

    def test_class_type(self):
        """Test the class type is correct."""
        sop = SOPClass("1.2.840.10008.5.1.4.1.1.2")
        assert isinstance(sop, SOPClass)
        assert isinstance(SOPClass(sop), SOPClass)

    def test_app_logging_sop(self):
        assert ProceduralEventLogging == "1.2.840.10008.1.40"
        assert (
            ProceduralEventLogging.service_class == ApplicationEventLoggingServiceClass
        )

    def test_basic_worklist_sop(self):
        """Test a Basic Worklist Service SOP Class."""
        assert ModalityWorklistInformationFind == "1.2.840.10008.5.1.4.31"
        assert (
            ModalityWorklistInformationFind.service_class
            == BasicWorklistManagementServiceClass
        )

    def test_color_palette_sop(self):
        """Test a Color Palette Service SOP Class."""
        assert ColorPaletteInformationModelMove == "1.2.840.10008.5.1.4.39.3"
        assert (
            ColorPaletteInformationModelMove.service_class
            == ColorPaletteQueryRetrieveServiceClass
        )

    def test_defined_procedure_sop(self):
        """Test a Defined Procedure Protocol Service SOP Class."""
        assert (
            DefinedProcedureProtocolInformationModelFind == "1.2.840.10008.5.1.4.20.1"
        )
        assert (
            DefinedProcedureProtocolInformationModelFind.service_class
            == DefinedProcedureProtocolQueryRetrieveServiceClass
        )

    def test_display_sop(self):
        assert DisplaySystem == "1.2.840.10008.5.1.1.40"
        assert DisplaySystem.service_class == DisplaySystemManagementServiceClass

    def test_hanging_protocol_sop(self):
        """Test a Hanging Protocol Service SOP Class."""
        assert HangingProtocolInformationModelGet == "1.2.840.10008.5.1.4.38.4"
        assert (
            HangingProtocolInformationModelGet.service_class
            == HangingProtocolQueryRetrieveServiceClass
        )

    def test_implant_template_sop(self):
        """Test an Implant Template Service SOP Class."""
        assert ImplantTemplateGroupInformationModelFind == "1.2.840.10008.5.1.4.45.2"
        assert (
            ImplantTemplateGroupInformationModelFind.service_class
            == ImplantTemplateQueryRetrieveServiceClass
        )

    def test_instance_sop(self):
        assert InstanceAvailabilityNotification == "1.2.840.10008.5.1.4.33"
        assert (
            InstanceAvailabilityNotification.service_class
            == InstanceAvailabilityNotificationServiceClass
        )

    def test_instance_sop(self):
        assert InventoryFind == "1.2.840.10008.5.1.4.1.1.201.2"
        assert InventoryFind.service_class == InventoryQueryRetrieveServiceClass

    def test_media_creation_sop(self):
        assert MediaCreationManagement == "1.2.840.10008.5.1.1.33"
        assert (
            MediaCreationManagement.service_class == MediaCreationManagementServiceClass
        )

    def test_media_storage_sop(self):
        assert MediaStorageDirectoryStorage == "1.2.840.10008.1.3.10"
        assert MediaStorageDirectoryStorage.service_class == ServiceClass

    def test_non_patient_sop(self):
        """Test a Non-Patient Object Service SOP Class."""
        assert HangingProtocolStorage == "1.2.840.10008.5.1.4.38.1"
        assert (
            HangingProtocolStorage.service_class == NonPatientObjectStorageServiceClass
        )

    def test_print_sop(self):
        assert PrintJob == "1.2.840.10008.5.1.1.14"
        assert PrintJob.service_class == PrintManagementServiceClass

    def test_procedure_step_sop(self):
        assert ModalityPerformedProcedureStep == "1.2.840.10008.3.1.2.3.3"
        assert ModalityPerformedProcedureStep.service_class == ProcedureStepServiceClass

    def test_protocol_approval_sop(self):
        """Test an Protocol Approval Service SOP Class."""
        assert ProtocolApprovalInformationModelFind == "1.2.840.10008.5.1.4.1.1.200.4"
        assert (
            ProtocolApprovalInformationModelFind.service_class
            == ProtocolApprovalQueryRetrieveServiceClass
        )

    def test_qr_sop(self):
        """Test a Query/Retrieve Service SOP Class."""
        assert (
            StudyRootQueryRetrieveInformationModelFind == "1.2.840.10008.5.1.4.1.2.2.1"
        )
        assert (
            StudyRootQueryRetrieveInformationModelFind.service_class
            == QueryRetrieveServiceClass
        )

    def test_relevant_patient_info_sop(self):
        """Test a Relevant Patient Information Query Service SOP Class."""
        assert GeneralRelevantPatientInformationQuery == "1.2.840.10008.5.1.4.37.1"
        assert (
            GeneralRelevantPatientInformationQuery.service_class
            == RelevantPatientInformationQueryServiceClass
        )

    def test_rt_sop(self):
        assert RTConventionalMachineVerification == "1.2.840.10008.5.1.4.34.8"
        assert (
            RTConventionalMachineVerification.service_class
            == RTMachineVerificationServiceClass
        )

    def test_storage_sop(self):
        """Test a Storage Service SOP Class."""
        assert CTImageStorage == "1.2.840.10008.5.1.4.1.1.2"
        assert CTImageStorage.service_class == StorageServiceClass

    def test_storage_commitment_sop(self):
        assert StorageCommitmentPushModel == "1.2.840.10008.1.20.1"
        assert StorageCommitmentPushModel.service_class == StorageCommitmentServiceClass

    def test_storage_management_sop(self):
        assert InventoryCreation == "1.2.840.10008.5.1.4.1.1.201.5"
        assert InventoryCreation.service_class == StorageManagementServiceClass

    def test_substance_admin_sop(self):
        """Test s Substance Administration Query Service SOP Class."""
        assert ProductCharacteristicsQuery == "1.2.840.10008.5.1.4.41"
        assert (
            ProductCharacteristicsQuery.service_class
            == SubstanceAdministrationQueryServiceClass
        )

    def test_ups_sop(self):
        assert UnifiedProcedureStepPull == "1.2.840.10008.5.1.4.34.6.3"
        assert (
            UnifiedProcedureStepPull.service_class == UnifiedProcedureStepServiceClass
        )

    def test_verification_sop(self):
        """Test a Verification Service SOP Class."""
        assert Verification == "1.2.840.10008.1.1"
        assert Verification.service_class == VerificationServiceClass

    def test_uid_creation(self):
        """Test creating a new UIDSOPClass."""
        sop = SOPClass("1.2.3")
        sop._service_class = ServiceClass

        assert sop == "1.2.3"
        assert sop.service_class == ServiceClass

        sop_b = SOPClass(sop)
        assert sop == sop_b
        assert sop_b == "1.2.3"
        assert sop_b.service_class == ServiceClass


class TestRegisterUID:
    def test_register_storage(self):
        """Test registering to the storage service."""
        register_uid(
            "1.2.3.4",
            "FooStorage",
            StorageServiceClass,
        )

        sop = sop_class.FooStorage
        assert sop == "1.2.3.4"
        assert sop.service_class == StorageServiceClass

        del _STORAGE_CLASSES["FooStorage"]
        delattr(sop_class, "FooStorage")

    def test_register_qr_find(self):
        """Test registering to the QR service - FIND."""
        register_uid(
            "1.2.3.4",
            "FooFind",
            QueryRetrieveServiceClass,
            dimse_msg_type="C-FIND",
        )

        sop = sop_class.FooFind
        assert sop == "1.2.3.4"
        assert sop.service_class == QueryRetrieveServiceClass
        assert sop in QueryRetrieveServiceClass._SUPPORTED_UIDS["C-FIND"]

        del _QR_CLASSES["FooFind"]
        QueryRetrieveServiceClass._SUPPORTED_UIDS["C-FIND"].remove(sop)
        delattr(sop_class, "FooFind")

    def test_register_qr_get(self):
        """Test registering to the QR service - GET."""
        register_uid(
            "1.2.3.4",
            "FooGet",
            QueryRetrieveServiceClass,
            dimse_msg_type="C-GET",
        )

        sop = sop_class.FooGet
        assert sop == "1.2.3.4"
        assert sop.service_class == QueryRetrieveServiceClass
        assert sop in QueryRetrieveServiceClass._SUPPORTED_UIDS["C-GET"]

        del _QR_CLASSES["FooGet"]
        QueryRetrieveServiceClass._SUPPORTED_UIDS["C-GET"].remove(sop)
        delattr(sop_class, "FooGet")

    def test_register_qr_move(self):
        """Test registering to the QR service - MOVE."""
        register_uid(
            "1.2.3.4",
            "FooMove",
            QueryRetrieveServiceClass,
            dimse_msg_type="C-MOVE",
        )

        sop = sop_class.FooMove
        assert sop == "1.2.3.4"
        assert sop.service_class == QueryRetrieveServiceClass
        assert sop in QueryRetrieveServiceClass._SUPPORTED_UIDS["C-MOVE"]

        del _QR_CLASSES["FooMove"]
        QueryRetrieveServiceClass._SUPPORTED_UIDS["C-MOVE"].remove(sop)
        delattr(sop_class, "FooMove")

    def test_register_bwm_find(self):
        """Test registering to the BWM service."""
        register_uid(
            "1.2.3.4",
            "FooFind",
            BasicWorklistManagementServiceClass,
            dimse_msg_type="C-FIND",
        )

        sop = sop_class.FooFind
        assert sop == "1.2.3.4"
        assert sop.service_class == BasicWorklistManagementServiceClass
        assert sop in BasicWorklistManagementServiceClass._SUPPORTED_UIDS["C-FIND"]

        del _BASIC_WORKLIST_CLASSES["FooFind"]
        BasicWorklistManagementServiceClass._SUPPORTED_UIDS["C-FIND"].remove(sop)
        delattr(sop_class, "FooFind")

    def test_register_substance_admin_find(self):
        """Test registering to the Substance Admin QR service."""
        register_uid(
            "1.2.3.4",
            "FooFind",
            SubstanceAdministrationQueryServiceClass,
            dimse_msg_type="C-FIND",
        )

        sop = sop_class.FooFind
        assert sop == "1.2.3.4"
        assert sop.service_class == SubstanceAdministrationQueryServiceClass
        assert sop in SubstanceAdministrationQueryServiceClass._SUPPORTED_UIDS["C-FIND"]

        del _SUBSTANCE_ADMINISTRATION_CLASSES["FooFind"]
        SubstanceAdministrationQueryServiceClass._SUPPORTED_UIDS["C-FIND"].remove(sop)
        delattr(sop_class, "FooFind")

    def test_invalid_keyword_raises(self):
        """Test invalid keyword raises exceptions."""
        msg = (
            "The keyword '2coo' is not a valid Python identifier or is "
            "a Python keyword"
        )
        with pytest.raises(ValueError, match=msg):
            register_uid("", "2coo", ServiceClass)

        msg = (
            "The keyword 'def' is not a valid Python identifier or is "
            "a Python keyword"
        )
        with pytest.raises(ValueError, match=msg):
            register_uid("", "def", ServiceClass)

    def test_invalid_service_class_raises(self):
        """Test that an invalid service_class raises exceptions."""
        msg = "'service_class' must be a class object not a class instance"
        with pytest.raises(TypeError, match=msg):
            register_uid("", "Foo", "")

        msg = (
            "'service_class' must be a ServiceClass subclass object "
            "such as 'StorageServiceClass'"
        )
        with pytest.raises(TypeError, match=msg):
            register_uid("", "Foo", str)

    def test_invalid_dimse_msg_type_raises(self):
        """Test that an invalid dimse_msg_type raises exceptions."""
        msg = (
            "'dimse_msg_type' must be 'C-FIND', 'C-GET' or 'C-MOVE' "
            "when registering a UID with QueryRetrieveServiceClass"
        )
        with pytest.raises(ValueError, match=msg):
            register_uid("", "Foo", QueryRetrieveServiceClass, "Foo")



"""Test DIMSE-N operations."""

try:
    from collections.abc import MutableSequence
except ImportError:
    from collections import MutableSequence
from io import BytesIO
import logging

import pytest

from pydicom import config as PYD_CONFIG
from pydicom.dataset import Dataset
from pydicom.dataelem import DataElement
from pydicom.tag import Tag, BaseTag
from pydicom.uid import UID

from pynetdicom import _config
from pynetdicom.dimse_messages import (
    N_EVENT_REPORT_RQ,
    N_EVENT_REPORT_RSP,
    N_GET_RQ,
    N_GET_RSP,
    N_SET_RQ,
    N_SET_RSP,
    N_ACTION_RQ,
    N_ACTION_RSP,
    N_CREATE_RQ,
    N_CREATE_RSP,
    N_DELETE_RQ,
    N_DELETE_RSP,
)
from pynetdicom.dimse_primitives import (
    N_EVENT_REPORT,
    N_GET,
    N_SET,
    N_ACTION,
    N_CREATE,
    N_DELETE,
)
from pynetdicom.utils import pretty_bytes
from pynetdicom.dsutils import decode, encode
from .encoded_dimse_n_msg import (
    n_er_rq_cmd,
    n_er_rq_ds,
    n_er_rsp_cmd,
    n_er_rsp_ds,
    n_get_rq_cmd,
    n_get_rsp_cmd,
    n_get_rsp_ds,
    n_delete_rq_cmd,
    n_delete_rsp_cmd,
    n_set_rq_cmd,
    n_set_rq_ds,
    n_set_rsp_cmd,
    n_set_rsp_ds,
    n_action_rq_cmd,
    n_action_rq_ds,
    n_action_rsp_cmd,
    n_action_rsp_ds,
    n_create_rq_cmd,
    n_create_rq_ds,
    n_create_rsp_cmd,
    n_create_rsp_ds,
)


if hasattr(PYD_CONFIG, "settings"):
    PYD_CONFIG.settings.reading_validation_mode = 0


LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.CRITICAL)


class TestPrimitive_N_EVENT:
    """Test DIMSE N-EVENT-REPORT operations."""

    def setup_method(self):
        self.default_conformance = _config.ENFORCE_UID_CONFORMANCE

    def teardown_method(self):
        _config.ENFORCE_UID_CONFORMANCE = self.default_conformance

    def test_assignment(self):
        """Check assignment works correctly"""
        primitive = N_EVENT_REPORT()

        # AffectedSOPClassUID
        primitive.AffectedSOPClassUID = "1.1.1"
        assert primitive.AffectedSOPClassUID == UID("1.1.1")
        assert isinstance(primitive.AffectedSOPClassUID, UID)
        primitive.AffectedSOPClassUID = UID("1.1.2")
        assert primitive.AffectedSOPClassUID == UID("1.1.2")
        assert isinstance(primitive.AffectedSOPClassUID, UID)
        primitive.AffectedSOPClassUID = b"1.1.3"
        assert primitive.AffectedSOPClassUID == UID("1.1.3")
        assert isinstance(primitive.AffectedSOPClassUID, UID)

        # AffectedSOPInstanceUID
        primitive.AffectedSOPInstanceUID = b"1.2.1"
        assert primitive.AffectedSOPInstanceUID == UID("1.2.1")
        assert isinstance(primitive.AffectedSOPClassUID, UID)
        primitive.AffectedSOPInstanceUID = UID("1.2.2")
        assert primitive.AffectedSOPInstanceUID == UID("1.2.2")
        assert isinstance(primitive.AffectedSOPClassUID, UID)
        primitive.AffectedSOPInstanceUID = "1.2.3"
        assert primitive.AffectedSOPInstanceUID == UID("1.2.3")
        assert isinstance(primitive.AffectedSOPClassUID, UID)

        # Event Information
        ds = Dataset()
        ds.PatientID = "1234567"
        primitive.EventInformation = BytesIO(encode(ds, True, True))
        ds = decode(primitive.EventInformation, True, True)
        assert ds.PatientID == "1234567"

        # Event Reply
        ds = Dataset()
        ds.PatientID = "123456"
        primitive.EventReply = BytesIO(encode(ds, True, True))
        ds = decode(primitive.EventReply, True, True)
        assert ds.PatientID == "123456"

        # Event Type ID
        primitive.EventTypeID = 0x0000
        assert primitive.EventTypeID == 0x0000

        # MessageID
        primitive.MessageID = 11
        assert 11 == primitive.MessageID

        # MessageIDBeingRespondedTo
        primitive.MessageIDBeingRespondedTo = 13
        assert 13 == primitive.MessageIDBeingRespondedTo

        # Status
        primitive.Status = 0x0000
        assert primitive.Status == 0x0000

    def test_uid_exceptions_false(self):
        """Test ValueError raised with ENFORCE_UID_CONFORMANCE = False."""
        primitive = N_EVENT_REPORT()

        _config.ENFORCE_UID_CONFORMANCE = False

        primitive.AffectedSOPClassUID = "abc"
        assert primitive.AffectedSOPClassUID == "abc"
        primitive.AffectedSOPInstanceUID = "abc"
        assert primitive.AffectedSOPInstanceUID == "abc"

        # Can't have more than 64 characters
        with pytest.raises(ValueError):
            primitive.AffectedSOPClassUID = "abc" * 22

        with pytest.raises(ValueError):
            primitive.AffectedSOPInstanceUID = "abc" * 22

    def test_uid_exceptions_true(self):
        """Test ValueError raised with ENFORCE_UID_CONFORMANCE = True."""
        primitive = N_EVENT_REPORT()
        _config.ENFORCE_UID_CONFORMANCE = True

        with pytest.raises(ValueError):
            primitive.AffectedSOPClassUID = "abc"

        with pytest.raises(ValueError):
            primitive.AffectedSOPInstanceUID = "abc"

    def test_exceptions(self):
        """Check incorrect types/values for properties raise exceptions"""
        primitive = N_EVENT_REPORT()

        # MessageID
        with pytest.raises(TypeError):
            primitive.MessageID = "halp"

        with pytest.raises(TypeError):
            primitive.MessageID = 1.111

        with pytest.raises(ValueError):
            primitive.MessageID = 65536

        with pytest.raises(ValueError):
            primitive.MessageID = -1

        # MessageIDBeingRespondedTo
        with pytest.raises(TypeError):
            primitive.MessageIDBeingRespondedTo = "halp"

        with pytest.raises(TypeError):
            primitive.MessageIDBeingRespondedTo = 1.111

        with pytest.raises(ValueError):
            primitive.MessageIDBeingRespondedTo = 65536

        with pytest.raises(ValueError):
            primitive.MessageIDBeingRespondedTo = -1

        # AffectedSOPClassUID
        with pytest.raises(TypeError):
            primitive.AffectedSOPClassUID = 45.2

        with pytest.raises(TypeError):
            primitive.AffectedSOPClassUID = 100

        # AffectedSOPInstanceUID
        with pytest.raises(TypeError):
            primitive.AffectedSOPInstanceUID = 45.2

        with pytest.raises(TypeError):
            primitive.AffectedSOPInstanceUID = 100

        # EventInformation
        msg = r"'EventInformation' parameter must be a BytesIO object"
        with pytest.raises(TypeError, match=msg):
            primitive.EventInformation = "halp"

        with pytest.raises(TypeError):
            primitive.EventInformation = 1.111

        with pytest.raises(TypeError):
            primitive.EventInformation = 50

        with pytest.raises(TypeError):
            primitive.EventInformation = [30, 10]

        # EventReply
        msg = r"'EventReply' parameter must be a BytesIO object"
        with pytest.raises(TypeError, match=msg):
            primitive.EventReply = "halp"

        with pytest.raises(TypeError):
            primitive.EventReply = 1.111

        with pytest.raises(TypeError):
            primitive.EventReply = 50

        with pytest.raises(TypeError):
            primitive.EventReply = [30, 10]

        # EventTypeID
        with pytest.raises(TypeError):
            primitive.EventTypeID = 19.4

        # Status
        with pytest.raises(TypeError):
            primitive.Status = 19.4

    def test_conversion_rq(self):
        """Check conversion to a -RQ PDU produces the correct output"""
        primitive = N_EVENT_REPORT()
        primitive.MessageID = 7
        primitive.AffectedSOPClassUID = "1.2.840.10008.5.1.4.1.1.2"
        primitive.AffectedSOPInstanceUID = "1.2.392.200036.9116.2.6.1.48"
        primitive.EventTypeID = 2

        ds = Dataset()
        ds.PatientID = "Test1101"
        ds.PatientName = "Tube HeNe"

        primitive.EventInformation = BytesIO(encode(ds, True, True))

        dimse_msg = N_EVENT_REPORT_RQ()
        dimse_msg.primitive_to_message(primitive)

        pdvs = []
        for fragment in dimse_msg.encode_msg(1, 16382):
            pdvs.append(fragment)

        assert len(pdvs) == 2
        cs_pdv = pdvs[0].presentation_data_value_list[0][1]
        ds_pdv = pdvs[1].presentation_data_value_list[0][1]

        assert cs_pdv == n_er_rq_cmd
        assert ds_pdv == n_er_rq_ds

    def test_conversion_rsp(self):
        """Check conversion to a -RSP PDU produces the correct output"""
        primitive = N_EVENT_REPORT()
        primitive.MessageIDBeingRespondedTo = 5
        primitive.AffectedSOPClassUID = "1.2.4.10"
        primitive.AffectedSOPInstanceUID = "1.2.4.5.7.8"
        primitive.Status = 0x0000
        primitive.EventTypeID = 2

        ds = Dataset()
        ds.PatientID = "Test1101"
        ds.PatientName = "Tube HeNe"

        primitive.EventReply = BytesIO(encode(ds, True, True))

        dimse_msg = N_EVENT_REPORT_RSP()
        dimse_msg.primitive_to_message(primitive)

        pdvs = []
        for fragment in dimse_msg.encode_msg(1, 16382):
            pdvs.append(fragment)

        assert len(pdvs) == 2
        cs_pdv = pdvs[0].presentation_data_value_list[0][1]
        ds_pdv = pdvs[1].presentation_data_value_list[0][1]

        assert cs_pdv == n_er_rsp_cmd
        assert ds_pdv == n_er_rsp_ds

    def test_is_valid_request(self):
        """Test N_EVENT_REPORT.is_valid_request"""
        primitive = N_EVENT_REPORT()
        assert not primitive.is_valid_request
        primitive.MessageID = 1
        assert not primitive.is_valid_request
        primitive.AffectedSOPClassUID = "1.2"
        assert not primitive.is_valid_request
        primitive.EventTypeID = 1
        assert not primitive.is_valid_request
        primitive.AffectedSOPInstanceUID = "1.2.1"
        assert primitive.is_valid_request

    def test_is_valid_resposne(self):
        """Test N_EVENT_REPORT.is_valid_response."""
        primitive = N_EVENT_REPORT()
        assert not primitive.is_valid_response
        primitive.MessageIDBeingRespondedTo = 1
        assert not primitive.is_valid_response
        primitive.Status = 0x0000
        assert primitive.is_valid_response


class TestPrimitive_N_GET:
    """Test DIMSE N-GET operations."""

    def setup_method(self):
        self.default_conformance = _config.ENFORCE_UID_CONFORMANCE

    def teardown_method(self):
        _config.ENFORCE_UID_CONFORMANCE = self.default_conformance

    def test_assignment(self):
        """Check assignment works correctly"""
        primitive = N_GET()

        # AffectedSOPClassUID
        primitive.AffectedSOPClassUID = "1.1.1"
        assert primitive.AffectedSOPClassUID == UID("1.1.1")
        assert isinstance(primitive.AffectedSOPClassUID, UID)
        primitive.AffectedSOPClassUID = UID("1.1.2")
        assert primitive.AffectedSOPClassUID == UID("1.1.2")
        assert isinstance(primitive.AffectedSOPClassUID, UID)
        primitive.AffectedSOPClassUID = b"1.1.3"
        assert primitive.AffectedSOPClassUID == UID("1.1.3")
        assert isinstance(primitive.AffectedSOPClassUID, UID)

        # AffectedSOPInstanceUID
        primitive.AffectedSOPInstanceUID = b"1.2.1"
        assert primitive.AffectedSOPInstanceUID == UID("1.2.1")
        assert isinstance(primitive.AffectedSOPClassUID, UID)
        primitive.AffectedSOPInstanceUID = UID("1.2.2")
        assert primitive.AffectedSOPInstanceUID == UID("1.2.2")
        assert isinstance(primitive.AffectedSOPClassUID, UID)
        primitive.AffectedSOPInstanceUID = "1.2.3"
        assert primitive.AffectedSOPInstanceUID == UID("1.2.3")
        assert isinstance(primitive.AffectedSOPClassUID, UID)

        # AttributeList
        ref_ds = Dataset()
        ref_ds.PatientID = "1234567"
        primitive.AttributeList = BytesIO(encode(ref_ds, True, True))
        ds = decode(primitive.AttributeList, True, True)
        assert ds.PatientID == "1234567"

        # AttributeIdentifierList
        primitive.AttributeIdentifierList = [
            0x00001000,
            (0x0000, 0x1000),
            Tag(0x7FE0, 0x0010),
        ]
        assert [
            Tag(0x0000, 0x1000),
            Tag(0x0000, 0x1000),
            Tag(0x7FE0, 0x0010),
        ] == primitive.AttributeIdentifierList
        primitive.AttributeIdentifierList = [(0x7FE0, 0x0010)]
        assert Tag(0x7FE0, 0x0010) == primitive.AttributeIdentifierList
        primitive.AttributeIdentifierList = (0x7FE0, 0x0010)
        assert Tag(0x7FE0, 0x0010) == primitive.AttributeIdentifierList

        elem = DataElement((0x0000, 0x0005), "AT", Tag(0x0000, 0x1000))
        assert isinstance(elem.value, BaseTag)
        primitive.AttributeIdentifierList = elem.value
        assert Tag(0x0000, 0x1000) == primitive.AttributeIdentifierList

        # MessageID
        primitive.MessageID = 11
        assert 11 == primitive.MessageID

        # MessageIDBeingRespondedTo
        primitive.MessageIDBeingRespondedTo = 13
        assert 13 == primitive.MessageIDBeingRespondedTo

        # RequestedSOPClassUID
        primitive.RequestedSOPClassUID = None
        assert primitive.RequestedSOPClassUID is None
        primitive.RequestedSOPClassUID = "1.1.1"
        assert primitive.RequestedSOPClassUID == UID("1.1.1")
        assert isinstance(primitive.RequestedSOPClassUID, UID)
        primitive.RequestedSOPClassUID = UID("1.1.2")
        assert primitive.RequestedSOPClassUID == UID("1.1.2")
        assert isinstance(primitive.RequestedSOPClassUID, UID)
        primitive.RequestedSOPClassUID = b"1.1.3"
        assert primitive.RequestedSOPClassUID == UID("1.1.3")
        assert isinstance(primitive.RequestedSOPClassUID, UID)

        # RequestedSOPInstanceUID
        primitive.RequestedSOPInstanceUID = None
        assert primitive.RequestedSOPInstanceUID is None
        primitive.RequestedSOPInstanceUID = b"1.2.1"
        assert primitive.RequestedSOPInstanceUID == UID("1.2.1")
        assert isinstance(primitive.RequestedSOPInstanceUID, UID)
        primitive.RequestedSOPInstanceUID = UID("1.2.2")
        assert primitive.RequestedSOPInstanceUID == UID("1.2.2")
        assert isinstance(primitive.RequestedSOPInstanceUID, UID)
        primitive.RequestedSOPInstanceUID = "1.2.3"
        assert primitive.RequestedSOPInstanceUID == UID("1.2.3")
        assert isinstance(primitive.RequestedSOPInstanceUID, UID)

        # Status
        primitive.Status = 0x0000
        assert primitive.Status == 0x0000

    def test_uid_exceptions_false(self):
        """Test ValueError raised with ENFORCE_UID_CONFORMANCE = False."""
        primitive = N_GET()

        _config.ENFORCE_UID_CONFORMANCE = False

        primitive.AffectedSOPClassUID = "abc"
        assert primitive.AffectedSOPClassUID == "abc"
        primitive.AffectedSOPInstanceUID = "abc"
        assert primitive.AffectedSOPInstanceUID == "abc"
        primitive.RequestedSOPClassUID = "abc"
        assert primitive.RequestedSOPClassUID == "abc"
        primitive.RequestedSOPInstanceUID = "abc"
        assert primitive.RequestedSOPInstanceUID == "abc"

        # Can't have more than 64 characters
        with pytest.raises(ValueError):
            primitive.AffectedSOPClassUID = "abc" * 22

        with pytest.raises(ValueError):
            primitive.AffectedSOPInstanceUID = "abc" * 22

        with pytest.raises(ValueError):
            primitive.RequestedSOPClassUID = "abc" * 22

        with pytest.raises(ValueError):
            primitive.RequestedSOPInstanceUID = "abc" * 22

    def test_uid_exceptions_true(self):
        """Test ValueError raised with ENFORCE_UID_CONFORMANCE = True."""
        primitive = N_GET()
        _config.ENFORCE_UID_CONFORMANCE = True

        with pytest.raises(ValueError):
            primitive.AffectedSOPClassUID = "abc"

        with pytest.raises(ValueError):
            primitive.AffectedSOPInstanceUID = "abc"

        with pytest.raises(ValueError):
            primitive.RequestedSOPClassUID = "abc"

        with pytest.raises(ValueError):
            primitive.RequestedSOPInstanceUID = "abc"

    def test_exceptions(self):
        """Check incorrect types/values for properties raise exceptions"""
        primitive = N_GET()

        # MessageID
        with pytest.raises(TypeError):
            primitive.MessageID = "halp"

        with pytest.raises(TypeError):
            primitive.MessageID = 1.111

        with pytest.raises(ValueError):
            primitive.MessageID = 65536

        with pytest.raises(ValueError):
            primitive.MessageID = -1

        # MessageIDBeingRespondedTo
        with pytest.raises(TypeError):
            primitive.MessageIDBeingRespondedTo = "halp"

        with pytest.raises(TypeError):
            primitive.MessageIDBeingRespondedTo = 1.111

        with pytest.raises(ValueError):
            primitive.MessageIDBeingRespondedTo = 65536

        with pytest.raises(ValueError):
            primitive.MessageIDBeingRespondedTo = -1

        # AffectedSOPClassUID
        with pytest.raises(TypeError):
            primitive.AffectedSOPClassUID = 45.2

        with pytest.raises(TypeError):
            primitive.AffectedSOPClassUID = 100

        # AffectedSOPInstanceUID
        with pytest.raises(TypeError):
            primitive.AffectedSOPInstanceUID = 45.2

        with pytest.raises(TypeError):
            primitive.AffectedSOPInstanceUID = 100

        # RequestedSOPClassUID
        with pytest.raises(TypeError):
            primitive.RequestedSOPClassUID = 45.2

        with pytest.raises(TypeError):
            primitive.RequestedSOPClassUID = 100

        # RequestedSOPInstanceUID
        with pytest.raises(TypeError):
            primitive.RequestedSOPInstanceUID = 45.2

        with pytest.raises(TypeError):
            primitive.RequestedSOPInstanceUID = 100

        # AttributeIdentifierList
        with pytest.raises(ValueError):
            primitive.AttributeIdentifierList = "ijk"

        with pytest.raises(ValueError, match="Attribute Identifier List must"):
            primitive.AttributeIdentifierList = ["ijk", "abc"]

        # AttributeList
        msg = r"'AttributeList' parameter must be a BytesIO object"
        with pytest.raises(TypeError, match=msg):
            primitive.AttributeList = "halp"

        with pytest.raises(TypeError):
            primitive.AttributeList = 1.111

        with pytest.raises(TypeError):
            primitive.AttributeList = 50

        with pytest.raises(TypeError):
            primitive.AttributeList = [30, 10]

        # Status
        with pytest.raises(TypeError):
            primitive.Status = 19.4

    def test_conversion_rq(self):
        """Check conversion to a -RQ PDU produces the correct output"""
        primitive = N_GET()
        primitive.MessageID = 7
        primitive.RequestedSOPClassUID = "1.2.840.10008.5.1.4.1.1.2"
        primitive.RequestedSOPInstanceUID = "1.2.392.200036.9116.2.6.1.48"
        primitive.AttributeIdentifierList = [
            (0x7FE0, 0x0010),
            (0x0000, 0x0000),
            (0xFFFF, 0xFFFF),
        ]

        dimse_msg = N_GET_RQ()
        dimse_msg.primitive_to_message(primitive)

        pdvs = []
        for fragment in dimse_msg.encode_msg(1, 16382):
            pdvs.append(fragment)

        assert len(pdvs) == 1
        cs_pdv = pdvs[0].presentation_data_value_list[0][1]

        assert cs_pdv == n_get_rq_cmd

    def test_conversion_rsp(self):
        """Check conversion to a -RSP PDU produces the correct output"""
        primitive = N_GET()
        primitive.MessageIDBeingRespondedTo = 5
        primitive.AffectedSOPClassUID = "1.2.4.10"
        primitive.AffectedSOPInstanceUID = "1.2.4.5.7.8"
        primitive.Status = 0x0000

        ds = Dataset()
        ds.PatientID = "Test1101"
        ds.PatientName = "Tube HeNe"

        primitive.AttributeList = BytesIO(encode(ds, True, True))

        dimse_msg = N_GET_RSP()
        dimse_msg.primitive_to_message(primitive)

        pdvs = []
        for fragment in dimse_msg.encode_msg(1, 16382):
            pdvs.append(fragment)

        assert len(pdvs) == 2
        cs_pdv = pdvs[0].presentation_data_value_list[0][1]
        ds_pdv = pdvs[1].presentation_data_value_list[0][1]

        assert cs_pdv == n_get_rsp_cmd
        assert ds_pdv == n_get_rsp_ds

    def test_is_valid_request(self):
        """Test N_GET.is_valid_request"""
        primitive = N_GET()
        assert not primitive.is_valid_request
        primitive.MessageID = 1
        assert not primitive.is_valid_request
        primitive.RequestedSOPClassUID = "1.2"
        assert not primitive.is_valid_request
        primitive.RequestedSOPInstanceUID = "1.2.1"
        assert primitive.is_valid_request

    def test_is_valid_resposne(self):
        """Test N_GET.is_valid_response."""
        primitive = N_GET()
        assert not primitive.is_valid_response
        primitive.MessageIDBeingRespondedTo = 1
        assert not primitive.is_valid_response
        primitive.Status = 0x0000
        assert primitive.is_valid_response


class TestPrimitive_N_SET:
    """Test DIMSE N-SET operations."""

    def setup_method(self):
        self.default_conformance = _config.ENFORCE_UID_CONFORMANCE

    def teardown_method(self):
        _config.ENFORCE_UID_CONFORMANCE = self.default_conformance

    def test_assignment(self):
        """Check assignment works correctly"""
        primitive = N_SET()

        # AffectedSOPClassUID
        primitive.AffectedSOPClassUID = "1.1.1"
        assert primitive.AffectedSOPClassUID == UID("1.1.1")
        assert isinstance(primitive.AffectedSOPClassUID, UID)
        primitive.AffectedSOPClassUID = UID("1.1.2")
        assert primitive.AffectedSOPClassUID == UID("1.1.2")
        assert isinstance(primitive.AffectedSOPClassUID, UID)
        primitive.AffectedSOPClassUID = b"1.1.3"
        assert primitive.AffectedSOPClassUID == UID("1.1.3")
        assert isinstance(primitive.AffectedSOPClassUID, UID)

        # AffectedSOPInstanceUID
        primitive.AffectedSOPInstanceUID = b"1.2.1"
        assert primitive.AffectedSOPInstanceUID == UID("1.2.1")
        assert isinstance(primitive.AffectedSOPClassUID, UID)
        primitive.AffectedSOPInstanceUID = UID("1.2.2")
        assert primitive.AffectedSOPInstanceUID == UID("1.2.2")
        assert isinstance(primitive.AffectedSOPClassUID, UID)
        primitive.AffectedSOPInstanceUID = "1.2.3"
        assert primitive.AffectedSOPInstanceUID == UID("1.2.3")
        assert isinstance(primitive.AffectedSOPClassUID, UID)

        # AttributeList
        ref_ds = Dataset()
        ref_ds.PatientID = "1234567"
        primitive.AttributeList = BytesIO(encode(ref_ds, True, True))
        ds = decode(primitive.AttributeList, True, True)
        assert ds.PatientID == "1234567"

        # ModificationList
        ref_ds = Dataset()
        ref_ds.PatientID = "123456"
        primitive.ModificationList = BytesIO(encode(ref_ds, True, True))
        ds = decode(primitive.ModificationList, True, True)
        assert ds.PatientID == "123456"

        # MessageID
        primitive.MessageID = 11
        assert 11 == primitive.MessageID

        # MessageIDBeingRespondedTo
        primitive.MessageIDBeingRespondedTo = 13
        assert 13 == primitive.MessageIDBeingRespondedTo

        # RequestedSOPClassUID
        primitive.RequestedSOPClassUID = "1.1.1"
        assert primitive.RequestedSOPClassUID == UID("1.1.1")
        assert isinstance(primitive.RequestedSOPClassUID, UID)
        primitive.RequestedSOPClassUID = UID("1.1.2")
        assert primitive.RequestedSOPClassUID == UID("1.1.2")
        assert isinstance(primitive.RequestedSOPClassUID, UID)
        primitive.RequestedSOPClassUID = b"1.1.3"
        assert primitive.RequestedSOPClassUID == UID("1.1.3")
        assert isinstance(primitive.RequestedSOPClassUID, UID)

        # RequestedSOPInstanceUID
        primitive.RequestedSOPInstanceUID = b"1.2.1"
        assert primitive.RequestedSOPInstanceUID == UID("1.2.1")
        assert isinstance(primitive.RequestedSOPInstanceUID, UID)
        primitive.RequestedSOPInstanceUID = UID("1.2.2")
        assert primitive.RequestedSOPInstanceUID == UID("1.2.2")
        assert isinstance(primitive.RequestedSOPInstanceUID, UID)
        primitive.RequestedSOPInstanceUID = "1.2.3"
        assert primitive.RequestedSOPInstanceUID == UID("1.2.3")
        assert isinstance(primitive.RequestedSOPInstanceUID, UID)

        # Status
        primitive.Status = 0x0000
        assert primitive.Status == 0x0000

    def test_uid_exceptions_false(self):
        """Test ValueError raised with ENFORCE_UID_CONFORMANCE = False."""
        primitive = N_SET()

        _config.ENFORCE_UID_CONFORMANCE = False

        primitive.AffectedSOPClassUID = "abc"
        assert primitive.AffectedSOPClassUID == "abc"
        primitive.AffectedSOPInstanceUID = "abc"
        assert primitive.AffectedSOPInstanceUID == "abc"
        primitive.RequestedSOPClassUID = "abc"
        assert primitive.RequestedSOPClassUID == "abc"
        primitive.RequestedSOPInstanceUID = "abc"
        assert primitive.RequestedSOPInstanceUID == "abc"

        # Can't have more than 64 characters
        with pytest.raises(ValueError):
            primitive.AffectedSOPClassUID = "abc" * 22

        with pytest.raises(ValueError):
            primitive.AffectedSOPInstanceUID = "abc" * 22

        with pytest.raises(ValueError):
            primitive.RequestedSOPClassUID = "abc" * 22

        with pytest.raises(ValueError):
            primitive.RequestedSOPInstanceUID = "abc" * 22

    def test_uid_exceptions_true(self):
        """Test ValueError raised with ENFORCE_UID_CONFORMANCE = True."""
        primitive = N_SET()
        _config.ENFORCE_UID_CONFORMANCE = True

        with pytest.raises(ValueError):
            primitive.AffectedSOPClassUID = "abc"

        with pytest.raises(ValueError):
            primitive.AffectedSOPInstanceUID = "abc"

        with pytest.raises(ValueError):
            primitive.RequestedSOPClassUID = "abc"

        with pytest.raises(ValueError):
            primitive.RequestedSOPInstanceUID = "abc"

    def test_exceptions(self):
        """Check incorrect types/values for properties raise exceptions"""
        primitive = N_SET()

        # MessageID
        with pytest.raises(TypeError):
            primitive.MessageID = "halp"

        with pytest.raises(TypeError):
            primitive.MessageID = 1.111

        with pytest.raises(ValueError):
            primitive.MessageID = 65536

        with pytest.raises(ValueError):
            primitive.MessageID = -1

        # MessageIDBeingRespondedTo
        with pytest.raises(TypeError):
            primitive.MessageIDBeingRespondedTo = "halp"

        with pytest.raises(TypeError):
            primitive.MessageIDBeingRespondedTo = 1.111

        with pytest.raises(ValueError):
            primitive.MessageIDBeingRespondedTo = 65536

        with pytest.raises(ValueError):
            primitive.MessageIDBeingRespondedTo = -1

        # AffectedSOPClassUID
        with pytest.raises(TypeError):
            primitive.AffectedSOPClassUID = 45.2

        with pytest.raises(TypeError):
            primitive.AffectedSOPClassUID = 100

        # AffectedSOPInstanceUID
        with pytest.raises(TypeError):
            primitive.AffectedSOPInstanceUID = 45.2

        with pytest.raises(TypeError):
            primitive.AffectedSOPInstanceUID = 100

        # RequestedSOPClassUID
        with pytest.raises(TypeError):
            primitive.RequestedSOPClassUID = 45.2

        with pytest.raises(TypeError):
            primitive.RequestedSOPClassUID = 100

        # RequestedSOPInstanceUID
        with pytest.raises(TypeError):
            primitive.RequestedSOPInstanceUID = 45.2

        with pytest.raises(TypeError):
            primitive.RequestedSOPInstanceUID = 100

        # AttributeList
        msg = r"'AttributeList' parameter must be a BytesIO object"
        with pytest.raises(TypeError, match=msg):
            primitive.AttributeList = "halp"

        with pytest.raises(TypeError):
            primitive.AttributeList = 1.111

        with pytest.raises(TypeError):
            primitive.AttributeList = 50

        with pytest.raises(TypeError):
            primitive.AttributeList = [30, 10]

        # ModificationList
        msg = r"'ModificationList' parameter must be a BytesIO object"
        with pytest.raises(TypeError, match=msg):
            primitive.ModificationList = "halp"

        with pytest.raises(TypeError):
            primitive.ModificationList = 1.111

        with pytest.raises(TypeError):
            primitive.ModificationList = 50

        with pytest.raises(TypeError):
            primitive.ModificationList = [30, 10]

        # Status
        with pytest.raises(TypeError):
            primitive.Status = 19.4

    def test_conversion_rq(self):
        """Check conversion to a -RQ PDU produces the correct output"""
        primitive = N_SET()
        primitive.MessageID = 7
        primitive.RequestedSOPClassUID = "1.2.840.10008.5.1.4.1.1.2"
        primitive.RequestedSOPInstanceUID = "1.2.392.200036.9116.2.6.1.48"

        ds = Dataset()
        ds.PatientID = "Test1101"
        ds.PatientName = "Tube HeNe"

        primitive.ModificationList = BytesIO(encode(ds, True, True))

        dimse_msg = N_SET_RQ()
        dimse_msg.primitive_to_message(primitive)

        pdvs = []
        for fragment in dimse_msg.encode_msg(1, 16382):
            pdvs.append(fragment)

        assert len(pdvs) == 2
        cs_pdv = pdvs[0].presentation_data_value_list[0][1]
        ds_pdv = pdvs[1].presentation_data_value_list[0][1]

        assert cs_pdv == n_set_rq_cmd
        assert ds_pdv == n_set_rq_ds

    def test_conversion_rsp(self):
        """Check conversion to a -RSP PDU produces the correct output"""
        primitive = N_SET()
        primitive.MessageIDBeingRespondedTo = 5
        primitive.AffectedSOPClassUID = "1.2.4.10"
        primitive.AffectedSOPInstanceUID = "1.2.4.5.7.8"
        primitive.Status = 0x0000

        ds = Dataset()
        ds.PatientID = "Test1101"
        ds.PatientName = "Tube HeNe"

        primitive.AttributeList = BytesIO(encode(ds, True, True))

        dimse_msg = N_SET_RSP()
        dimse_msg.primitive_to_message(primitive)

        pdvs = []
        for fragment in dimse_msg.encode_msg(1, 16382):
            pdvs.append(fragment)

        assert len(pdvs) == 2
        cs_pdv = pdvs[0].presentation_data_value_list[0][1]
        ds_pdv = pdvs[1].presentation_data_value_list[0][1]

        assert cs_pdv == n_set_rsp_cmd
        assert ds_pdv == n_set_rsp_ds

    def test_is_valid_request(self):
        """Test N_SET.is_valid_request"""
        primitive = N_SET()
        assert not primitive.is_valid_request
        primitive.MessageID = 1
        assert not primitive.is_valid_request
        primitive.RequestedSOPClassUID = "1.2"
        assert not primitive.is_valid_request
        primitive.RequestedSOPInstanceUID = "1.2.1"
        assert not primitive.is_valid_request
        ds = Dataset()
        ds.PatientID = "Test1101"
        ds.PatientName = "Tube HeNe"

        primitive.ModificationList = BytesIO(encode(ds, True, True))
        assert primitive.is_valid_request

    def test_is_valid_resposne(self):
        """Test N_SET.is_valid_response."""
        primitive = N_SET()
        assert not primitive.is_valid_response
        primitive.MessageIDBeingRespondedTo = 1
        assert not primitive.is_valid_response
        primitive.Status = 0x0000
        assert primitive.is_valid_response


class TestPrimitive_N_ACTION:
    """Test DIMSE N-ACTION operations."""

    def setup_method(self):
        self.default_conformance = _config.ENFORCE_UID_CONFORMANCE

    def teardown_method(self):
        _config.ENFORCE_UID_CONFORMANCE = self.default_conformance

    def test_assignment(self):
        """Check assignment works correctly"""
        primitive = N_ACTION()

        # Action Type ID
        primitive.ActionTypeID = 0x0000
        assert primitive.ActionTypeID == 0x0000

        # AffectedSOPClassUID
        primitive.AffectedSOPClassUID = "1.1.1"
        assert primitive.AffectedSOPClassUID == UID("1.1.1")
        assert isinstance(primitive.AffectedSOPClassUID, UID)
        primitive.AffectedSOPClassUID = UID("1.1.2")
        assert primitive.AffectedSOPClassUID == UID("1.1.2")
        assert isinstance(primitive.AffectedSOPClassUID, UID)
        primitive.AffectedSOPClassUID = b"1.1.3"
        assert primitive.AffectedSOPClassUID == UID("1.1.3")
        assert isinstance(primitive.AffectedSOPClassUID, UID)

        # AffectedSOPInstanceUID
        primitive.AffectedSOPInstanceUID = b"1.2.1"
        assert primitive.AffectedSOPInstanceUID == UID("1.2.1")
        assert isinstance(primitive.AffectedSOPClassUID, UID)
        primitive.AffectedSOPInstanceUID = UID("1.2.2")
        assert primitive.AffectedSOPInstanceUID == UID("1.2.2")
        assert isinstance(primitive.AffectedSOPClassUID, UID)
        primitive.AffectedSOPInstanceUID = "1.2.3"
        assert primitive.AffectedSOPInstanceUID == UID("1.2.3")
        assert isinstance(primitive.AffectedSOPClassUID, UID)

        # ActionInformation
        ds = Dataset()
        ds.PatientID = "1234567"
        primitive.ActionInformation = BytesIO(encode(ds, True, True))
        ds = decode(primitive.ActionInformation, True, True)
        assert ds.PatientID == "1234567"

        # ActionReply
        ds = Dataset()
        ds.PatientID = "123456"
        primitive.ActionReply = BytesIO(encode(ds, True, True))
        ds = decode(primitive.ActionReply, True, True)
        assert ds.PatientID == "123456"

        # MessageID
        primitive.MessageID = 11
        assert 11 == primitive.MessageID

        # MessageIDBeingRespondedTo
        primitive.MessageIDBeingRespondedTo = 13
        assert 13 == primitive.MessageIDBeingRespondedTo

        # RequestedSOPClassUID
        primitive.RequestedSOPClassUID = "1.1.1"
        assert primitive.RequestedSOPClassUID == UID("1.1.1")
        assert isinstance(primitive.RequestedSOPClassUID, UID)
        primitive.RequestedSOPClassUID = UID("1.1.2")
        assert primitive.RequestedSOPClassUID == UID("1.1.2")
        assert isinstance(primitive.RequestedSOPClassUID, UID)
        primitive.RequestedSOPClassUID = b"1.1.3"
        assert primitive.RequestedSOPClassUID == UID("1.1.3")
        assert isinstance(primitive.RequestedSOPClassUID, UID)

        # RequestedSOPInstanceUID
        primitive.RequestedSOPInstanceUID = b"1.2.1"
        assert primitive.RequestedSOPInstanceUID == UID("1.2.1")
        assert isinstance(primitive.RequestedSOPInstanceUID, UID)
        primitive.RequestedSOPInstanceUID = UID("1.2.2")
        assert primitive.RequestedSOPInstanceUID == UID("1.2.2")
        assert isinstance(primitive.RequestedSOPInstanceUID, UID)
        primitive.RequestedSOPInstanceUID = "1.2.3"
        assert primitive.RequestedSOPInstanceUID == UID("1.2.3")
        assert isinstance(primitive.RequestedSOPInstanceUID, UID)

        # Status
        primitive.Status = 0x0000
        assert primitive.Status == 0x0000

    def test_uid_exceptions_false(self):
        """Test ValueError raised with ENFORCE_UID_CONFORMANCE = False."""
        primitive = N_ACTION()

        _config.ENFORCE_UID_CONFORMANCE = False

        primitive.AffectedSOPClassUID = "abc"
        assert primitive.AffectedSOPClassUID == "abc"
        primitive.AffectedSOPInstanceUID = "abc"
        assert primitive.AffectedSOPInstanceUID == "abc"
        primitive.RequestedSOPClassUID = "abc"
        assert primitive.RequestedSOPClassUID == "abc"
        primitive.RequestedSOPInstanceUID = "abc"
        assert primitive.RequestedSOPInstanceUID == "abc"

        # Can't have more than 64 characters
        with pytest.raises(ValueError):
            primitive.AffectedSOPClassUID = "abc" * 22

        with pytest.raises(ValueError):
            primitive.AffectedSOPInstanceUID = "abc" * 22

        with pytest.raises(ValueError):
            primitive.RequestedSOPClassUID = "abc" * 22

        with pytest.raises(ValueError):
            primitive.RequestedSOPInstanceUID = "abc" * 22

    def test_uid_exceptions_true(self):
        """Test ValueError raised with ENFORCE_UID_CONFORMANCE = True."""
        primitive = N_ACTION()
        _config.ENFORCE_UID_CONFORMANCE = True

        with pytest.raises(ValueError):
            primitive.AffectedSOPClassUID = "abc"

        with pytest.raises(ValueError):
            primitive.AffectedSOPInstanceUID = "abc"

        with pytest.raises(ValueError):
            primitive.RequestedSOPClassUID = "abc"

        with pytest.raises(ValueError):
            primitive.RequestedSOPInstanceUID = "abc"

    def test_exceptions(self):
        """Check incorrect types/values for properties raise exceptions"""
        primitive = N_ACTION()

        # MessageID
        with pytest.raises(TypeError):
            primitive.MessageID = "halp"

        with pytest.raises(TypeError):
            primitive.MessageID = 1.111

        with pytest.raises(ValueError):
            primitive.MessageID = 65536

        with pytest.raises(ValueError):
            primitive.MessageID = -1

        # MessageIDBeingRespondedTo
        with pytest.raises(TypeError):
            primitive.MessageIDBeingRespondedTo = "halp"

        with pytest.raises(TypeError):
            primitive.MessageIDBeingRespondedTo = 1.111

        with pytest.raises(ValueError):
            primitive.MessageIDBeingRespondedTo = 65536

        with pytest.raises(ValueError):
            primitive.MessageIDBeingRespondedTo = -1

        # AffectedSOPClassUID
        with pytest.raises(TypeError):
            primitive.AffectedSOPClassUID = 45.2

        with pytest.raises(TypeError):
            primitive.AffectedSOPClassUID = 100

        # AffectedSOPInstanceUID
        with pytest.raises(TypeError):
            primitive.AffectedSOPInstanceUID = 45.2

        with pytest.raises(TypeError):
            primitive.AffectedSOPInstanceUID = 100

        # RequestedSOPClassUID
        with pytest.raises(TypeError):
            primitive.RequestedSOPClassUID = 45.2

        with pytest.raises(TypeError):
            primitive.RequestedSOPClassUID = 100

        # RequestedSOPInstanceUID
        with pytest.raises(TypeError):
            primitive.RequestedSOPInstanceUID = 45.2

        with pytest.raises(TypeError):
            primitive.RequestedSOPInstanceUID = 100

        # ActionInformation
        msg = r"'ActionInformation' parameter must be a BytesIO object"
        with pytest.raises(TypeError, match=msg):
            primitive.ActionInformation = "halp"

        with pytest.raises(TypeError):
            primitive.ActionInformation = 1.111

        with pytest.raises(TypeError):
            primitive.ActionInformation = 50

        with pytest.raises(TypeError):
            primitive.ActionInformation = [30, 10]

        # ActionReply
        msg = r"'ActionReply' parameter must be a BytesIO object"
        with pytest.raises(TypeError, match=msg):
            primitive.ActionReply = "halp"

        with pytest.raises(TypeError):
            primitive.ActionReply = 1.111

        with pytest.raises(TypeError):
            primitive.ActionReply = 50

        with pytest.raises(TypeError):
            primitive.ActionReply = [30, 10]

        # ActionTypeID
        with pytest.raises(TypeError):
            primitive.ActionTypeID = 19.4

        # Status
        with pytest.raises(TypeError):
            primitive.Status = 19.4

    def test_conversion_rq(self):
        """Check conversion to a -RQ PDU produces the correct output"""
        primitive = N_ACTION()
        primitive.MessageID = 7
        primitive.RequestedSOPClassUID = "1.2.840.10008.5.1.4.1.1.2"
        primitive.RequestedSOPInstanceUID = "1.2.392.200036.9116.2.6.1.48"
        primitive.ActionTypeID = 1

        ds = Dataset()
        ds.PatientID = "Test1101"
        ds.PatientName = "Tube HeNe"

        primitive.ActionInformation = BytesIO(encode(ds, True, True))

        dimse_msg = N_ACTION_RQ()
        dimse_msg.primitive_to_message(primitive)

        pdvs = []
        for fragment in dimse_msg.encode_msg(1, 16382):
            pdvs.append(fragment)

        assert len(pdvs) == 2
        cs_pdv = pdvs[0].presentation_data_value_list[0][1]
        ds_pdv = pdvs[1].presentation_data_value_list[0][1]

        assert cs_pdv == n_action_rq_cmd
        assert ds_pdv == n_action_rq_ds

    def test_conversion_rsp(self):
        """Check conversion to a -RSP PDU produces the correct output"""
        primitive = N_ACTION()
        primitive.MessageIDBeingRespondedTo = 5
        primitive.AffectedSOPClassUID = "1.2.4.10"
        primitive.AffectedSOPInstanceUID = "1.2.4.5.7.8"
        primitive.Status = 0x0000
        primitive.ActionTypeID = 1

        ds = Dataset()
        ds.PatientID = "Test1101"
        ds.PatientName = "Tube HeNe"

        primitive.ActionReply = BytesIO(encode(ds, True, True))

        dimse_msg = N_ACTION_RSP()
        dimse_msg.primitive_to_message(primitive)

        pdvs = []
        for fragment in dimse_msg.encode_msg(1, 16382):
            pdvs.append(fragment)

        assert len(pdvs) == 2
        cs_pdv = pdvs[0].presentation_data_value_list[0][1]
        ds_pdv = pdvs[1].presentation_data_value_list[0][1]

        assert cs_pdv == n_action_rsp_cmd
        assert ds_pdv == n_action_rsp_ds

    def test_is_valid_request(self):
        """Test N_ACTION.is_valid_request"""
        primitive = N_ACTION()
        assert not primitive.is_valid_request
        primitive.MessageID = 1
        assert not primitive.is_valid_request
        primitive.RequestedSOPClassUID = "1.2"
        assert not primitive.is_valid_request
        primitive.RequestedSOPInstanceUID = "1.2.1"
        assert not primitive.is_valid_request
        primitive.ActionTypeID = 4
        assert primitive.is_valid_request

    def test_is_valid_resposne(self):
        """Test N_ACTION.is_valid_response."""
        primitive = N_ACTION()
        assert not primitive.is_valid_response
        primitive.MessageIDBeingRespondedTo = 1
        assert not primitive.is_valid_response
        primitive.Status = 0x0000
        assert primitive.is_valid_response


class TestPrimitive_N_CREATE:
    """Test DIMSE N-CREATE operations."""

    def setup_method(self):
        self.default_conformance = _config.ENFORCE_UID_CONFORMANCE

    def teardown_method(self):
        _config.ENFORCE_UID_CONFORMANCE = self.default_conformance

    def test_assignment(self):
        """Check assignment works correctly"""
        primitive = N_CREATE()

        # AffectedSOPClassUID
        primitive.AffectedSOPClassUID = "1.1.1"
        assert primitive.AffectedSOPClassUID == UID("1.1.1")
        assert isinstance(primitive.AffectedSOPClassUID, UID)
        primitive.AffectedSOPClassUID = UID("1.1.2")
        assert primitive.AffectedSOPClassUID == UID("1.1.2")
        assert isinstance(primitive.AffectedSOPClassUID, UID)
        primitive.AffectedSOPClassUID = b"1.1.3"
        assert primitive.AffectedSOPClassUID == UID("1.1.3")
        assert isinstance(primitive.AffectedSOPClassUID, UID)

        # AffectedSOPInstanceUID
        primitive.AffectedSOPInstanceUID = b"1.2.1"
        assert primitive.AffectedSOPInstanceUID == UID("1.2.1")
        assert isinstance(primitive.AffectedSOPClassUID, UID)
        primitive.AffectedSOPInstanceUID = UID("1.2.2")
        assert primitive.AffectedSOPInstanceUID == UID("1.2.2")
        assert isinstance(primitive.AffectedSOPClassUID, UID)
        primitive.AffectedSOPInstanceUID = "1.2.3"
        assert primitive.AffectedSOPInstanceUID == UID("1.2.3")
        assert isinstance(primitive.AffectedSOPClassUID, UID)

        # AttributeList
        ref_ds = Dataset()
        ref_ds.PatientID = "1234567"
        primitive.AttributeList = BytesIO(encode(ref_ds, True, True))
        ds = decode(primitive.AttributeList, True, True)
        assert ds.PatientID == "1234567"

        # MessageID
        primitive.MessageID = 11
        assert 11 == primitive.MessageID

        # MessageIDBeingRespondedTo
        primitive.MessageIDBeingRespondedTo = 13
        assert 13 == primitive.MessageIDBeingRespondedTo

        # Status
        primitive.Status = 0x0000
        assert primitive.Status == 0x0000

    def test_uid_exceptions_false(self):
        """Test ValueError raised with ENFORCE_UID_CONFORMANCE = False."""
        primitive = N_CREATE()

        _config.ENFORCE_UID_CONFORMANCE = False

        primitive.AffectedSOPClassUID = "abc"
        assert primitive.AffectedSOPClassUID == "abc"
        primitive.AffectedSOPInstanceUID = "abc"
        assert primitive.AffectedSOPInstanceUID == "abc"

        # Can't have more than 64 characters
        with pytest.raises(ValueError):
            primitive.AffectedSOPClassUID = "abc" * 22

        with pytest.raises(ValueError):
            primitive.AffectedSOPInstanceUID = "abc" * 22

    def test_uid_exceptions_true(self):
        """Test ValueError raised with ENFORCE_UID_CONFORMANCE = True."""
        primitive = N_CREATE()
        _config.ENFORCE_UID_CONFORMANCE = True

        with pytest.raises(ValueError):
            primitive.AffectedSOPClassUID = "abc"

        with pytest.raises(ValueError):
            primitive.AffectedSOPInstanceUID = "abc"

    def test_exceptions(self):
        """Check incorrect types/values for properties raise exceptions"""
        primitive = N_CREATE()

        # MessageID
        with pytest.raises(TypeError):
            primitive.MessageID = "halp"

        with pytest.raises(TypeError):
            primitive.MessageID = 1.111

        with pytest.raises(ValueError):
            primitive.MessageID = 65536

        with pytest.raises(ValueError):
            primitive.MessageID = -1

        # MessageIDBeingRespondedTo
        with pytest.raises(TypeError):
            primitive.MessageIDBeingRespondedTo = "halp"

        with pytest.raises(TypeError):
            primitive.MessageIDBeingRespondedTo = 1.111

        with pytest.raises(ValueError):
            primitive.MessageIDBeingRespondedTo = 65536

        with pytest.raises(ValueError):
            primitive.MessageIDBeingRespondedTo = -1

        # AffectedSOPClassUID
        with pytest.raises(TypeError):
            primitive.AffectedSOPClassUID = 45.2

        with pytest.raises(TypeError):
            primitive.AffectedSOPClassUID = 100

        # AffectedSOPInstanceUID
        with pytest.raises(TypeError):
            primitive.AffectedSOPInstanceUID = 45.2

        with pytest.raises(TypeError):
            primitive.AffectedSOPInstanceUID = 100

        # AttributeList
        msg = r"'AttributeList' parameter must be a BytesIO object"
        with pytest.raises(TypeError, match=msg):
            primitive.AttributeList = "halp"

        with pytest.raises(TypeError):
            primitive.AttributeList = 1.111

        with pytest.raises(TypeError):
            primitive.AttributeList = 50

        with pytest.raises(TypeError):
            primitive.AttributeList = [30, 10]

        # Status
        with pytest.raises(TypeError):
            primitive.Status = 19.4

    def test_conversion_rq(self):
        """Check conversion to a -RQ PDU produces the correct output."""
        primitive = N_CREATE()
        primitive.MessageID = 7
        primitive.AffectedSOPClassUID = "1.2.840.10008.5.1.4.1.1.2"
        primitive.AffectedSOPInstanceUID = "1.2.392.200036.9116.2.6.1.48"

        ds = Dataset()
        ds.PatientID = "Test1101"
        ds.PatientName = "Tube HeNe"

        primitive.AttributeList = BytesIO(encode(ds, True, True))

        dimse_msg = N_CREATE_RQ()
        dimse_msg.primitive_to_message(primitive)

        pdvs = []
        for fragment in dimse_msg.encode_msg(1, 16382):
            pdvs.append(fragment)

        assert len(pdvs) == 2
        cs_pdv = pdvs[0].presentation_data_value_list[0][1]
        ds_pdv = pdvs[1].presentation_data_value_list[0][1]

        assert cs_pdv == n_create_rq_cmd
        assert ds_pdv == n_create_rq_ds

    def test_conversion_rsp(self):
        """Check conversion to a -RSP PDU produces the correct output"""
        primitive = N_CREATE()
        primitive.MessageIDBeingRespondedTo = 5
        primitive.AffectedSOPClassUID = "1.2.4.10"
        primitive.AffectedSOPInstanceUID = "1.2.4.5.7.8"
        primitive.Status = 0x0000

        ds = Dataset()
        ds.PatientID = "Test1101"
        ds.PatientName = "Tube HeNe"

        primitive.AttributeList = BytesIO(encode(ds, True, True))

        dimse_msg = N_CREATE_RSP()
        dimse_msg.primitive_to_message(primitive)

        pdvs = []
        for fragment in dimse_msg.encode_msg(1, 16382):
            pdvs.append(fragment)

        assert len(pdvs) == 2
        cs_pdv = pdvs[0].presentation_data_value_list[0][1]
        ds_pdv = pdvs[1].presentation_data_value_list[0][1]

        assert cs_pdv == n_create_rsp_cmd
        assert ds_pdv == n_create_rsp_ds

    def test_is_valid_request(self):
        """Test N_CREATE.is_valid_request"""
        primitive = N_CREATE()
        assert not primitive.is_valid_request
        primitive.MessageID = 1
        assert not primitive.is_valid_request
        primitive.AffectedSOPClassUID = "1.2"
        assert primitive.is_valid_request

    def test_is_valid_response(self):
        """Test N_CREATE.is_valid_response."""
        primitive = N_CREATE()
        assert not primitive.is_valid_response
        primitive.MessageIDBeingRespondedTo = 1
        assert not primitive.is_valid_response
        primitive.Status = 0x0000
        assert primitive.is_valid_response


class TestPrimitive_N_DELETE:
    """Test DIMSE N-DELETE operations."""

    def setup_method(self):
        self.default_conformance = _config.ENFORCE_UID_CONFORMANCE

    def teardown_method(self):
        _config.ENFORCE_UID_CONFORMANCE = self.default_conformance

    def test_assignment(self):
        """Check assignment works correctly"""
        primitive = N_DELETE()

        # AffectedSOPClassUID
        primitive.AffectedSOPClassUID = "1.1.1"
        assert primitive.AffectedSOPClassUID == UID("1.1.1")
        assert isinstance(primitive.AffectedSOPClassUID, UID)
        primitive.AffectedSOPClassUID = UID("1.1.2")
        assert primitive.AffectedSOPClassUID == UID("1.1.2")
        assert isinstance(primitive.AffectedSOPClassUID, UID)
        primitive.AffectedSOPClassUID = b"1.1.3"
        assert primitive.AffectedSOPClassUID == UID("1.1.3")
        assert isinstance(primitive.AffectedSOPClassUID, UID)

        # AffectedSOPInstanceUID
        primitive.AffectedSOPInstanceUID = b"1.2.1"
        assert primitive.AffectedSOPInstanceUID == UID("1.2.1")
        assert isinstance(primitive.AffectedSOPClassUID, UID)
        primitive.AffectedSOPInstanceUID = UID("1.2.2")
        assert primitive.AffectedSOPInstanceUID == UID("1.2.2")
        assert isinstance(primitive.AffectedSOPClassUID, UID)
        primitive.AffectedSOPInstanceUID = "1.2.3"
        assert primitive.AffectedSOPInstanceUID == UID("1.2.3")
        assert isinstance(primitive.AffectedSOPClassUID, UID)

        # MessageID
        primitive.MessageID = 11
        assert 11 == primitive.MessageID

        # MessageIDBeingRespondedTo
        primitive.MessageIDBeingRespondedTo = 13
        assert 13 == primitive.MessageIDBeingRespondedTo

        # RequestedSOPClassUID
        primitive.RequestedSOPClassUID = "1.1.1"
        assert primitive.RequestedSOPClassUID == UID("1.1.1")
        assert isinstance(primitive.RequestedSOPClassUID, UID)
        primitive.RequestedSOPClassUID = UID("1.1.2")
        assert primitive.RequestedSOPClassUID == UID("1.1.2")
        assert isinstance(primitive.RequestedSOPClassUID, UID)
        primitive.RequestedSOPClassUID = b"1.1.3"
        assert primitive.RequestedSOPClassUID == UID("1.1.3")
        assert isinstance(primitive.RequestedSOPClassUID, UID)

        # RequestedSOPInstanceUID
        primitive.RequestedSOPInstanceUID = b"1.2.1"
        assert primitive.RequestedSOPInstanceUID == UID("1.2.1")
        assert isinstance(primitive.RequestedSOPInstanceUID, UID)
        primitive.RequestedSOPInstanceUID = UID("1.2.2")
        assert primitive.RequestedSOPInstanceUID == UID("1.2.2")
        assert isinstance(primitive.RequestedSOPInstanceUID, UID)
        primitive.RequestedSOPInstanceUID = "1.2.3"
        assert primitive.RequestedSOPInstanceUID == UID("1.2.3")
        assert isinstance(primitive.RequestedSOPInstanceUID, UID)

        # Status
        primitive.Status = 0x0000
        assert primitive.Status == 0x0000

    def test_uid_exceptions_false(self):
        """Test ValueError raised with ENFORCE_UID_CONFORMANCE = False."""
        primitive = N_DELETE()

        _config.ENFORCE_UID_CONFORMANCE = False

        primitive.AffectedSOPClassUID = "abc"
        assert primitive.AffectedSOPClassUID == "abc"
        primitive.AffectedSOPInstanceUID = "abc"
        assert primitive.AffectedSOPInstanceUID == "abc"
        primitive.RequestedSOPClassUID = "abc"
        assert primitive.RequestedSOPClassUID == "abc"
        primitive.RequestedSOPInstanceUID = "abc"
        assert primitive.RequestedSOPInstanceUID == "abc"

        # Can't have more than 64 characters
        with pytest.raises(ValueError):
            primitive.AffectedSOPClassUID = "abc" * 22

        with pytest.raises(ValueError):
            primitive.AffectedSOPInstanceUID = "abc" * 22

        with pytest.raises(ValueError):
            primitive.RequestedSOPClassUID = "abc" * 22

        with pytest.raises(ValueError):
            primitive.RequestedSOPInstanceUID = "abc" * 22

    def test_uid_exceptions_true(self):
        """Test ValueError raised with ENFORCE_UID_CONFORMANCE = True."""
        primitive = N_DELETE()
        _config.ENFORCE_UID_CONFORMANCE = True

        with pytest.raises(ValueError):
            primitive.AffectedSOPClassUID = "abc"

        with pytest.raises(ValueError):
            primitive.AffectedSOPInstanceUID = "abc"

        with pytest.raises(ValueError):
            primitive.RequestedSOPClassUID = "abc"

        with pytest.raises(ValueError):
            primitive.RequestedSOPInstanceUID = "abc"

    def test_exceptions(self):
        """Check incorrect types/values for properties raise exceptions"""
        primitive = N_DELETE()

        # MessageID
        with pytest.raises(TypeError):
            primitive.MessageID = "halp"

        with pytest.raises(TypeError):
            primitive.MessageID = 1.111

        with pytest.raises(ValueError):
            primitive.MessageID = 65536

        with pytest.raises(ValueError):
            primitive.MessageID = -1

        # MessageIDBeingRespondedTo
        with pytest.raises(TypeError):
            primitive.MessageIDBeingRespondedTo = "halp"

        with pytest.raises(TypeError):
            primitive.MessageIDBeingRespondedTo = 1.111

        with pytest.raises(ValueError):
            primitive.MessageIDBeingRespondedTo = 65536

        with pytest.raises(ValueError):
            primitive.MessageIDBeingRespondedTo = -1

        # AffectedSOPClassUID
        with pytest.raises(TypeError):
            primitive.AffectedSOPClassUID = 45.2

        with pytest.raises(TypeError):
            primitive.AffectedSOPClassUID = 100

        # AffectedSOPInstanceUID
        with pytest.raises(TypeError):
            primitive.AffectedSOPInstanceUID = 45.2

        with pytest.raises(TypeError):
            primitive.AffectedSOPInstanceUID = 100

        # RequestedSOPClassUID
        with pytest.raises(TypeError):
            primitive.RequestedSOPClassUID = 45.2

        with pytest.raises(TypeError):
            primitive.RequestedSOPClassUID = 100

        # RequestedSOPInstanceUID
        with pytest.raises(TypeError):
            primitive.RequestedSOPInstanceUID = 45.2

        with pytest.raises(TypeError):
            primitive.RequestedSOPInstanceUID = 100

        # Status
        with pytest.raises(TypeError):
            primitive.Status = 19.4

    def test_conversion_rq(self):
        """Check conversion to a -RQ PDU produces the correct output"""
        primitive = N_DELETE()
        primitive.MessageID = 7
        primitive.RequestedSOPClassUID = "1.2.3"
        primitive.RequestedSOPInstanceUID = "1.2.30"

        dimse_msg = N_DELETE_RQ()
        dimse_msg.primitive_to_message(primitive)

        pdvs = []
        for fragment in dimse_msg.encode_msg(1, 16382):
            pdvs.append(fragment)

        assert len(pdvs) == 1
        cs_pdv = pdvs[0].presentation_data_value_list[0][1]

        assert cs_pdv == n_delete_rq_cmd

    def test_conversion_rsp(self):
        """Check conversion to a -RSP PDU produces the correct output"""
        primitive = N_DELETE()
        primitive.MessageIDBeingRespondedTo = 5
        primitive.AffectedSOPClassUID = "1.2.4.10"
        primitive.AffectedSOPInstanceUID = "1.2.4.5.7.8"
        primitive.Status = 0xC201

        dimse_msg = N_DELETE_RSP()
        dimse_msg.primitive_to_message(primitive)

        pdvs = []
        for fragment in dimse_msg.encode_msg(1, 16382):
            pdvs.append(fragment)

        assert len(pdvs) == 1
        cs_pdv = pdvs[0].presentation_data_value_list[0][1]

        assert cs_pdv == n_delete_rsp_cmd

    def test_is_valid_request(self):
        """Test N_DELETE.is_valid_request"""
        primitive = N_DELETE()
        assert not primitive.is_valid_request
        primitive.MessageID = 1
        assert not primitive.is_valid_request
        primitive.RequestedSOPClassUID = "1.2"
        assert not primitive.is_valid_request
        primitive.RequestedSOPInstanceUID = "1.2.1"
        assert primitive.is_valid_request

    def test_is_valid_resposne(self):
        """Test N_DELETE.is_valid_response."""
        primitive = N_DELETE()
        assert not primitive.is_valid_response
        primitive.MessageIDBeingRespondedTo = 1
        assert not primitive.is_valid_response
        primitive.Status = 0x0000
        assert primitive.is_valid_response



"""Tests for the NonPatientObjectStorageServiceClass."""

from io import BytesIO
import os

import pytest

from pydicom import dcmread
from pydicom.dataset import Dataset
from pydicom.uid import ExplicitVRLittleEndian

from pynetdicom import AE, evt, debug_logger
from pynetdicom.dimse_primitives import C_STORE
from pynetdicom.sop_class import (
    HangingProtocolStorage,
)


# debug_logger()

TEST_DS_DIR = os.path.join(os.path.dirname(__file__), "dicom_files")
DATASET = dcmread(os.path.join(TEST_DS_DIR, "CTImageStorage.dcm"))
DATASET.SOPClassUID = HangingProtocolStorage


class TestNonPatientObjectStorageServiceClass:
    """Test the NonPatientObjectStorageServiceClass.

    Subclass of StorageServiceClass with its own set of statuses.
    """

    def setup_method(self):
        """Run prior to each test"""
        self.ae = None

    def teardown_method(self):
        """Clear any active threads"""
        if self.ae:
            self.ae.shutdown()

    def test_scp_failed_ds_decode(self):
        """Test failure to decode the dataset"""

        # Hard to test directly as decode errors won't show up until the
        #   dataset is actually used
        def handle(event):
            try:
                for elem in event.dataset.iterall():
                    pass
            except:
                status = Dataset()
                status.Status = 0xC210
                status.ErrorComment = "Unable to decode the dataset"
                return status

            return 0x0000

        handlers = [(evt.EVT_C_STORE, handle)]

        self.ae = ae = AE()
        ae.add_supported_context(HangingProtocolStorage)
        ae.add_requested_context(HangingProtocolStorage, ExplicitVRLittleEndian)
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)

        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        req = C_STORE()
        req.MessageID = 1
        req.AffectedSOPClassUID = DATASET.SOPClassUID
        req.AffectedSOPInstanceUID = DATASET.SOPInstanceUID
        req.Priority = 0x0002
        # Bad VR? AA
        req.DataSet = BytesIO(
            b"\x08\x00\x01\x00\x40\x40\x00\x00\x00\x00\x00\x08\x00\x49"
        )

        # Send C-STORE request to DIMSE and get response
        # Need to manually hit the checkpoint
        assoc._reactor_checkpoint.clear()
        assoc.dimse.send_msg(req, 1)
        with pytest.warns(UserWarning):
            cx_id, rsp = assoc.dimse.get_msg(True)
        assoc._reactor_checkpoint.set()

        assert rsp.Status == 0xC210
        assert rsp.ErrorComment == "Unable to decode the dataset"
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_scp_handler_return_dataset(self):
        """Test handler returning a Dataset status"""

        def handle(event):
            status = Dataset()
            status.Status = 0x0001
            return status

        handlers = [(evt.EVT_C_STORE, handle)]

        self.ae = ae = AE()
        ae.add_supported_context(HangingProtocolStorage)
        ae.add_requested_context(HangingProtocolStorage)
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)

        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established
        rsp = assoc.send_c_store(DATASET)
        assert rsp.Status == 0x0001
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_scp_handler_return_dataset_multi(self):
        """Test handler returning a Dataset status with other elements"""

        def handle(event):
            status = Dataset()
            status.Status = 0x0001
            status.ErrorComment = "Test"
            status.OffendingElement = 0x00080010
            return status

        handlers = [(evt.EVT_C_STORE, handle)]

        self.ae = ae = AE()
        ae.add_supported_context(HangingProtocolStorage)
        ae.add_requested_context(HangingProtocolStorage)
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)

        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established
        rsp = assoc.send_c_store(DATASET)
        assert rsp.Status == 0x0001
        assert rsp.ErrorComment == "Test"
        assert rsp.OffendingElement == 0x00080010
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_scp_handler_return_int(self):
        """Test handler returning an int status"""

        def handle(event):
            return 0x0000

        handlers = [(evt.EVT_C_STORE, handle)]

        self.ae = ae = AE()
        ae.add_supported_context(HangingProtocolStorage)
        ae.add_requested_context(HangingProtocolStorage)
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)

        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established
        rsp = assoc.send_c_store(DATASET)
        assert rsp.Status == 0x0000
        assert "ErrorComment" not in rsp
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_scp_handler_return_invalid(self):
        """Test handler returning an invalid status"""

        def handle(event):
            return 0xFFF0

        handlers = [(evt.EVT_C_STORE, handle)]

        self.ae = ae = AE()
        ae.add_supported_context(HangingProtocolStorage)
        ae.add_requested_context(HangingProtocolStorage)
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)

        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established
        rsp = assoc.send_c_store(DATASET)
        assert rsp.Status == 0xFFF0
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_scp_handler_no_status(self):
        """Test handler not returning a status"""

        def handle(event):
            return None

        handlers = [(evt.EVT_C_STORE, handle)]

        self.ae = ae = AE()
        ae.add_supported_context(HangingProtocolStorage)
        ae.add_requested_context(HangingProtocolStorage)
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)

        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established
        rsp = assoc.send_c_store(DATASET)
        assert rsp.Status == 0xC002
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_scp_handler_exception(self):
        """Test handler raising an exception"""

        def handle(event):
            raise ValueError

        handlers = [(evt.EVT_C_STORE, handle)]

        self.ae = ae = AE()
        ae.add_supported_context(HangingProtocolStorage)
        ae.add_requested_context(HangingProtocolStorage)
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)

        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established
        rsp = assoc.send_c_store(DATASET)
        assert rsp.Status == 0xC211
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_scp_handler_context(self):
        """Test handler event's context attribute"""
        attrs = {}

        def handle(event):
            attrs["context"] = event.context
            attrs["assoc"] = event.assoc
            attrs["request"] = event.request
            attrs["dataset"] = event.dataset
            return 0x0000

        handlers = [(evt.EVT_C_STORE, handle)]

        self.ae = ae = AE()
        ae.add_supported_context(HangingProtocolStorage)
        ae.add_requested_context(HangingProtocolStorage)
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)

        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established
        status = assoc.send_c_store(DATASET)
        assert status.Status == 0x0000
        assoc.release()
        assert assoc.is_released

        cx = attrs["context"]
        assert cx.context_id == 1
        assert cx.abstract_syntax == HangingProtocolStorage
        assert cx.transfer_syntax == "1.2.840.10008.1.2"

        scp.shutdown()

    def test_scp_handler_assoc(self):
        """Test handler event's assoc attribute"""
        attrs = {}

        def handle(event):
            attrs["context"] = event.context
            attrs["assoc"] = event.assoc
            attrs["request"] = event.request
            attrs["dataset"] = event.dataset
            return 0x0000

        handlers = [(evt.EVT_C_STORE, handle)]

        self.ae = ae = AE()
        ae.add_supported_context(HangingProtocolStorage)
        ae.add_requested_context(HangingProtocolStorage)
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)

        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established
        status = assoc.send_c_store(DATASET)
        assert status.Status == 0x0000

        scp_assoc = attrs["assoc"]
        assert scp_assoc == scp.active_associations[0]

        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_scp_handler_request(self):
        """Test handler event's request attribute"""
        attrs = {}

        def handle(event):
            attrs["context"] = event.context
            attrs["assoc"] = event.assoc
            attrs["request"] = event.request
            attrs["dataset"] = event.dataset
            return 0x0000

        handlers = [(evt.EVT_C_STORE, handle)]

        self.ae = ae = AE()
        ae.add_supported_context(HangingProtocolStorage)
        ae.add_requested_context(HangingProtocolStorage)
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)

        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established
        status = assoc.send_c_store(DATASET)
        assert status.Status == 0x0000
        assoc.release()
        assert assoc.is_released

        req = attrs["request"]
        assert req.MessageID == 1
        assert isinstance(req, C_STORE)

        scp.shutdown()

    def test_scp_handler_dataset(self):
        """Test handler event's dataset property"""
        attrs = {}

        def handle(event):
            attrs["context"] = event.context
            attrs["assoc"] = event.assoc
            attrs["request"] = event.request
            attrs["dataset"] = event.dataset
            return 0x0000

        handlers = [(evt.EVT_C_STORE, handle)]

        self.ae = ae = AE()
        ae.add_supported_context(HangingProtocolStorage)
        ae.add_requested_context(HangingProtocolStorage)
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)

        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established
        status = assoc.send_c_store(DATASET)
        assert status.Status == 0x0000
        assoc.release()
        assert assoc.is_released

        ds = attrs["dataset"]
        assert isinstance(ds, Dataset)
        assert ds.PatientName == DATASET.PatientName

        scp.shutdown()

    def test_scp_handler_move_origin(self):
        """Test handler event's request property with MoveOriginator"""
        attrs = {}

        def handle(event):
            attrs["context"] = event.context
            attrs["assoc"] = event.assoc
            attrs["request"] = event.request
            attrs["dataset"] = event.dataset
            return 0x0000

        handlers = [(evt.EVT_C_STORE, handle)]

        self.ae = ae = AE()
        ae.add_supported_context(HangingProtocolStorage)
        ae.add_requested_context(HangingProtocolStorage)
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)

        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established
        status = assoc.send_c_store(DATASET, originator_aet="ORIGIN", originator_id=888)
        assert status.Status == 0x0000
        assoc.release()
        assert assoc.is_released

        req = attrs["request"]
        assert req.MoveOriginatorApplicationEntityTitle == "ORIGIN"
        assert req.MoveOriginatorMessageID == 888

        scp.shutdown()



#!/usr/bin/env python
"""An Echo SCU application.

Used for verifying basic DICOM connectivity and as such has a focus on
providing useful debugging and logging information.
"""

import argparse
import sys

from pydicom.uid import (
    ExplicitVRLittleEndian,
    ImplicitVRLittleEndian,
    ExplicitVRBigEndian,
    DeflatedExplicitVRLittleEndian,
)

from pynetdicom import AE
from pynetdicom.apps.common import setup_logging
from pynetdicom._globals import DEFAULT_MAX_LENGTH
from pynetdicom.sop_class import Verification


__version__ = "0.7.0"


def _setup_argparser():
    """Setup the command line arguments"""
    # Description
    parser = argparse.ArgumentParser(
        description=(
            "The echoscu application implements a Service Class User "
            "(SCU) for the Verification Service Class. It sends a DICOM "
            "C-ECHO message to a Service Class Provider (SCP) and "
            "waits for a response. The application can be used to "
            "verify basic DICOM connectivity."
        ),
        usage="echoscu [options] addr port",
    )

    # Parameters
    req_opts = parser.add_argument_group("Parameters")
    req_opts.add_argument("addr", help="TCP/IP address of DICOM peer", type=str)
    req_opts.add_argument("port", help="TCP/IP port number of peer", type=int)

    # General Options
    gen_opts = parser.add_argument_group("General Options")
    gen_opts.add_argument(
        "--version", help="print version information and exit", action="store_true"
    )
    output = gen_opts.add_mutually_exclusive_group()
    output.add_argument(
        "-q",
        "--quiet",
        help="quiet mode, print no warnings and errors",
        action="store_const",
        dest="log_type",
        const="q",
    )
    output.add_argument(
        "-v",
        "--verbose",
        help="verbose mode, print processing details",
        action="store_const",
        dest="log_type",
        const="v",
    )
    output.add_argument(
        "-d",
        "--debug",
        help="debug mode, print debug information",
        action="store_const",
        dest="log_type",
        const="d",
    )
    gen_opts.add_argument(
        "-ll",
        "--log-level",
        metavar="[l]",
        help=("use level l for the logger (critical, error, warn, info, debug)"),
        type=str,
        choices=["critical", "error", "warn", "info", "debug"],
    )

    # Network Options
    net_opts = parser.add_argument_group("Network Options")
    net_opts.add_argument(
        "-aet",
        "--calling-aet",
        metavar="[a]etitle",
        help="set my calling AE title (default: ECHOSCU)",
        type=str,
        default="ECHOSCU",
    )
    net_opts.add_argument(
        "-aec",
        "--called-aet",
        metavar="[a]etitle",
        help="set called AE title of peer (default: ANY-SCP)",
        type=str,
        default="ANY-SCP",
    )
    net_opts.add_argument(
        "-ta",
        "--acse-timeout",
        metavar="[s]econds",
        help="timeout for ACSE messages (default: 30 s)",
        type=float,
        default=30,
    )
    net_opts.add_argument(
        "-td",
        "--dimse-timeout",
        metavar="[s]econds",
        help="timeout for DIMSE messages (default: 30 s)",
        type=float,
        default=30,
    )
    net_opts.add_argument(
        "-tn",
        "--network-timeout",
        metavar="[s]econds",
        help="timeout for the network (default: 30 s)",
        type=float,
        default=30,
    )
    net_opts.add_argument(
        "-pdu",
        "--max-pdu",
        metavar="[n]umber of bytes",
        help=(
            f"set max receive pdu to n bytes (0 for unlimited, "
            f"default: {DEFAULT_MAX_LENGTH})"
        ),
        type=int,
        default=DEFAULT_MAX_LENGTH,
    )

    # Transfer Syntaxes
    ts_opts = parser.add_argument_group("Transfer Syntax Options")
    syntax = ts_opts.add_mutually_exclusive_group()
    syntax.add_argument(
        "-xe",
        "--request-little",
        help="request explicit VR little endian TS only",
        action="store_true",
    )
    syntax.add_argument(
        "-xb",
        "--request-big",
        help="request explicit VR big endian TS only",
        action="store_true",
    )
    syntax.add_argument(
        "-xi",
        "--request-implicit",
        help="request implicit VR little endian TS only",
        action="store_true",
    )

    # Miscellaneous Options
    misc_opts = parser.add_argument_group("Miscellaneous Options")
    misc_opts.add_argument(
        "--repeat",
        metavar="[n]umber",
        help="repeat echo request n times",
        type=int,
        default=1,
    )
    misc_opts.add_argument(
        "--abort", help="abort association instead of releasing it", action="store_true"
    )

    return parser.parse_args()


def main(args=None):
    """Run the application."""
    if args is not None:
        sys.argv = args

    args = _setup_argparser()

    if args.version:
        print(f"echoscu.py v{__version__}")
        sys.exit()

    APP_LOGGER = setup_logging(args, "echoscu")
    APP_LOGGER.debug("echoscu.py v%s", __version__)
    APP_LOGGER.debug("")

    # Set Transfer Syntax options
    transfer_syntax = [
        ExplicitVRLittleEndian,
        ImplicitVRLittleEndian,
        DeflatedExplicitVRLittleEndian,
        ExplicitVRBigEndian,
    ]

    if args.request_little:
        transfer_syntax = [ExplicitVRLittleEndian]
    elif args.request_big:
        transfer_syntax = [ExplicitVRBigEndian]
    elif args.request_implicit:
        transfer_syntax = [ImplicitVRLittleEndian]

    # Create local AE
    ae = AE(ae_title=args.calling_aet)
    ae.add_requested_context(Verification, transfer_syntax)

    # Set timeouts
    ae.acse_timeout = args.acse_timeout
    ae.dimse_timeout = args.dimse_timeout
    ae.network_timeout = args.network_timeout

    # Request association with remote AE
    assoc = ae.associate(
        args.addr, args.port, ae_title=args.called_aet, max_pdu=args.max_pdu
    )

    # If we successfully associated then send C-ECHO
    if assoc.is_established:
        for ii in range(args.repeat):
            assoc.send_c_echo()

        # Abort or release association
        if args.abort:
            assoc.abort()
        else:
            assoc.release()
    else:
        # Failed to associate: timeout, refused, connection closed, aborted
        sys.exit(1)


if __name__ == "__main__":
    main()



"""The Parrot testing server."""

import gc
import logging
import select
import socket
from socketserver import TCPServer, ThreadingMixIn, BaseRequestHandler
import struct
from struct import pack, unpack
import threading
import time

from pynetdicom.transport import AssociationServer


LOGGER = logging.getLogger(__name__)


class ParrotRequest(BaseRequestHandler):
    @property
    def commands(self):
        """Return a list of command tuples."""
        # commands: recv, send, exit
        return self.server.commands

    @property
    def event(self):
        return self.server.event

    def handle(self):
        # self.server: ThreadedParrot
        # self.client_address: remote's (host, port)
        # self.request: socket
        self.event.clear()
        self.received = []
        self.sent = []
        for cmd, data in self.commands:
            if cmd == "recv":
                self.kill_read = False
                while not self.kill_read:
                    if self.ready:
                        self.received.append(bytes(self.read_data))
                        self.kill_read = True
            elif cmd == "send":
                self.send(data)
                self.sent.append(data)
            elif cmd == "wait":
                time.sleep(data)

        # Disconnects automatically when this method ends!

    @property
    def ready(self):
        """Return True if there is data available to be read.

        Returns
        -------
        bool
            True if the socket has data ready to be read, False otherwise.
        """
        try:
            # Use a timeout of 0 so we get an "instant" result
            ready, _, _ = select.select([self.socket], [], [], 0.5)
        except (OSError, TimeoutError, ValueError):
            return False

        return bool(ready)

    @property
    def read_data(self):
        bytestream = bytearray()

        # Try and read the PDU type and length from the socket
        try:
            bytestream.extend(self.recv(6))
        except (OSError, TimeoutError):
            pass
        try:
            # Byte 1 is always the PDU type
            # Byte 2 is always reserved
            # Bytes 3-6 are always the PDU length
            pdu_type, _, pdu_length = unpack(">BBL", bytestream)
        except struct.error:
            pass

        # Try and read the rest of the PDU
        try:
            bytestream += self.recv(pdu_length)
        except (OSError, TimeoutError):
            pass

        return bytestream

    def recv(self, nr_bytes):
        """Read `nr_bytes` from the socket.

        *Events Emitted*

        - None

        Parameters
        ----------
        nr_bytes : int
            The number of bytes to attempt to read from the socket.

        Returns
        -------
        bytearray
            The data read from the socket.
        """
        bytestream = bytearray()
        nr_read = 0
        # socket.recv() returns when the network buffer has been emptied
        #   not necessarily when the number of bytes requested have been
        #   read. Its up to us to keep calling recv() until we have all the
        #   data we want
        # **BLOCKING** until either all the data is read or an error occurs
        while nr_read < nr_bytes:
            # Python docs recommend reading a relatively small power of 2
            #   such as 4096
            bufsize = 4096
            if (nr_bytes - nr_read) < bufsize:
                bufsize = nr_bytes - nr_read

            bytes_read = self.socket.recv(bufsize)

            # If socket.recv() reads 0 bytes then the connection has been
            #   broken, so return what we have so far
            if not bytes_read:
                return bytestream

            bytestream.extend(bytes_read)
            nr_read += len(bytes_read)

        return bytestream

    def send(self, bytestream):
        """Try and send `bystream` to the remote.

        Parameters
        ----------
        bytestream : bytes
            The data to send to the remote.

        Returns
        -------
        bool
            True if the data was sent successfully, False otherwise.
        """
        total_sent = 0
        length_data = len(bytestream)
        try:
            while total_sent < length_data:
                # Returns the number of bytes sent
                nr_sent = self.socket.send(bytestream)
                total_sent += nr_sent
        except (OSError, TimeoutError):
            # Evt17: Transport connection closed
            return False

        return True

    @property
    def socket(self):
        """Return the socket to the remote."""
        return self.request


class SteppingParrotRequest(ParrotRequest):
    def handle(self):
        # self.server: ThreadedParrot
        # self.client_address: remote's (host, port)
        # self.request: socket
        self.event.clear()
        self.received = []
        self.sent = []
        while True:
            LOGGER.debug("Parrot: waiting on event.wait()")
            self.event.wait()
            LOGGER.debug("Parrot: event.wait() ended")
            cmd, data = self.commands.pop(0)
            LOGGER.debug(f"Parrot: running command '{cmd}'")

            if cmd == "exit":
                LOGGER.debug("   Parrot: exiting...")
                return
            elif cmd == "recv":
                LOGGER.debug("   Parrot: receiving data")
                self.kill_read = False
                while not self.kill_read:
                    if self.ready:
                        self.received.append(bytes(self.read_data))
                        self.kill_read = True
                LOGGER.debug(f"  Parrot: received {len(self.received)} bytes")
            elif cmd == "send":
                LOGGER.debug("   Parrot: sending data")
                self.send(data)
                self.sent.append(data)

            LOGGER.debug("Parrot: step ending, event.clear()")
            self.event.clear()


def start_server(commands, handler=SteppingParrotRequest):
    server = ThreadedParrot(("localhost", 11112), commands, handler)
    thread = threading.Thread(target=server.serve_forever)
    thread.daemon = True
    thread.start()

    return server


class Parrot(AssociationServer):
    def __init__(self, address, commands, handler=SteppingParrotRequest):
        """Create a new AssociationServer, bind a socket and start listening.

        Parameters
        ----------
        ae : ae.ApplicationEntity
            The parent AE that's running the server.
        address : 2-tuple
            The (host, port) that the server should run on.
        ssl_context : ssl.SSLContext, optional
            If TLS is to be used then this should be the ssl.SSLContext used
            to wrap the client sockets, otherwise if None then no TLS will be
            used (default).
        """
        self.commands = commands
        self.event = threading.Event()
        self.event.set()
        self.ssl_context = None
        self.allow_reuse_address = True

        TCPServer.__init__(self, address, handler, bind_and_activate=True)

        self.timeout = 60
        self.handlers = []
        self._gc = [0, 59]

    @property
    def received(self):
        return self.handlers[0].received

    def server_bind(self):
        """Bind the socket and set the socket options.

        - socket.SO_REUSEADDR is set to 1
        - socket.SO_RCVTIMEO is set to AE.network_timeout unless the value is
          None in which case it will be left unset.
        """
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_RCVTIMEO, pack("ll", 2, 0))

        self.socket.bind(self.server_address)
        self.server_address = self.socket.getsockname()

    def step(self):
        while self.event.is_set():
            time.sleep(0.0001)

        self.event.set()

    def service_actions(self) -> None:
        """Called by the serve_forever() loop"""
        # For whatever reason dead Association threads aren't being garbage
        #   collected so do it manually when a request is received
        if self._gc[0] == self._gc[1]:
            gc.collect()
            self._gc[0] = 0
            return

        self._gc[0] += 1

    def finish_request(self, request, client_address):
        """Finish one request by instantiating RequestHandlerClass."""
        handler = self.RequestHandlerClass(request, client_address, self)
        self.handlers.append(handler)

    def shutdown(self):
        """Completely shutdown the server and close it's socket."""
        TCPServer.shutdown(self)
        self.server_close()

    process_request = TCPServer.process_request


class ThreadedParrot(ThreadingMixIn, Parrot):
    process_request_thread = ThreadingMixIn.process_request_thread



"""Implementation of the DIMSE Status values."""

from enum import IntEnum

from pydicom.dataset import Dataset

from pynetdicom._globals import (
    STATUS_SUCCESS,
    STATUS_FAILURE,
    STATUS_WARNING,
    STATUS_CANCEL,
    STATUS_PENDING,
    STATUS_UNKNOWN,
)


StatusDictType = dict[int, tuple[str, str]]


# Non-Service Class specific statuses - PS3.7 Annex C
GENERAL_STATUS: StatusDictType = {
    0x0000: (STATUS_SUCCESS, ""),
    0x0105: (STATUS_FAILURE, "No Such Attribute"),
    0x0106: (STATUS_FAILURE, "Invalid Attribute Value"),
    0x0107: (STATUS_WARNING, "Attribute List Error"),
    0x0110: (STATUS_FAILURE, "Processing Failure"),
    0x0111: (STATUS_FAILURE, "Duplication SOP Instance"),
    0x0112: (STATUS_FAILURE, "No Such SOP Instance"),
    0x0113: (STATUS_FAILURE, "No Such Event Type"),
    0x0114: (STATUS_FAILURE, "No Such Argument"),
    0x0115: (STATUS_FAILURE, "Invalid Argument Value"),
    0x0116: (STATUS_WARNING, "Attribute Value Out of Range"),
    0x0117: (STATUS_FAILURE, "Invalid Object Instance"),
    0x0118: (STATUS_FAILURE, "No Such SOP Class"),
    0x0119: (STATUS_FAILURE, "Class-Instance Conflict"),
    0x0120: (STATUS_FAILURE, "Missing Attribute"),
    0x0121: (STATUS_FAILURE, "Missing Attribute Value"),
    0x0122: (STATUS_FAILURE, "Refused: SOP Class Not Supported"),
    0x0123: (STATUS_FAILURE, "No Such Action"),
    0x0124: (STATUS_FAILURE, "Refused: Not Authorised"),
    0x0210: (STATUS_FAILURE, "Duplicate Invocation"),
    0x0211: (STATUS_FAILURE, "Unrecognised Operation"),
    0x0212: (STATUS_FAILURE, "Mistyped Argument"),
    0x0213: (STATUS_FAILURE, "Resource Limitation"),
    0xFE00: (STATUS_CANCEL, ""),
}


# SERVICE CLASS STATUSES
# Verification Service Class specific status code values
VERIFICATION_SERVICE_CLASS_STATUS = GENERAL_STATUS


# Storage Service Class specific status code values - PS3.4 Annex B.2.3
STORAGE_SERVICE_CLASS_STATUS: StatusDictType = {
    0xB000: (STATUS_WARNING, "Coercion of Data Elements"),
    0xB007: (STATUS_WARNING, "Data Set Does Not Match SOP Class"),
    0xB006: (STATUS_WARNING, "Element Discarded"),
}

# Ranged values
for _code in range(0xA700, 0xA7FF + 1):
    STORAGE_SERVICE_CLASS_STATUS[_code] = (STATUS_FAILURE, "Refused: Out of Resources")
for _code in range(0xA900, 0xA9FF + 1):
    STORAGE_SERVICE_CLASS_STATUS[_code] = (
        STATUS_FAILURE,
        "Data Set Does Not Match SOP Class",
    )
for _code in range(0xC000, 0xCFFF + 1):
    STORAGE_SERVICE_CLASS_STATUS[_code] = (STATUS_FAILURE, "Cannot Understand")

# Add the General status code values - PS3.7 9.1.1.1.9 and Annex C
STORAGE_SERVICE_CLASS_STATUS.update(GENERAL_STATUS)


# Query Retrieve - FIND specific status code values
#   PS3.4 Annex C.4.1.1.4 and Annex C.6.4.3
# Hanging Protocol - FIND specific status code values
#   PS3.4 Annex U.4.1
# Color Palette - FIND specific status code values
#   PS3.4 Annex X
# Implant Template - FIND specific status code values
#   PS3.4 Annex BB
# Defined Procedure Protocol - FIND specific status code values
#   PS3.4 Annex HH
QR_FIND_SERVICE_CLASS_STATUS: StatusDictType = {
    0xA700: (STATUS_FAILURE, "Refused: Out of Resources"),
    0xA710: (STATUS_FAILURE, "Invalid Prior Record Key"),
    0xA900: (STATUS_FAILURE, "Identifier Does Not Match SOP Class"),
    0xB001: (
        STATUS_WARNING,
        "Matching reached response limit, subsequent request may return additional matches",
    ),
    0xFF00: (STATUS_PENDING, "Matches are continuing, current match supplied"),
    0xFF01: (STATUS_PENDING, "Matches are continuing, warning"),
}

# Ranged values
for _code in range(0xC000, 0xCFFF + 1):
    QR_FIND_SERVICE_CLASS_STATUS[_code] = (STATUS_FAILURE, "Unable to Process")

# Add the General status code values - PS3.7 Annex C
QR_FIND_SERVICE_CLASS_STATUS.update(GENERAL_STATUS)


# Query Retrieve - MOVE specific status code values
#   PS3.4 Annex C.4.2.1.5 and Annex Y.4.1.1.4
# Hanging Protocol - MOVE specific status code values
#   PS3.4 Annex U.4.2
# Color Palette - MOVE specific status code values
#   PS3.4 Annex X
# Implant Template - MOVE specific status code values
#   PS3.4 Annex BB
# Defined Procedure Protocol - MOVE specific status code values
#   PS3.4 Annex HH
QR_MOVE_SERVICE_CLASS_STATUS: StatusDictType = {
    0xA701: (
        STATUS_FAILURE,
        "Refused: Out of resources, unable to calculate number of matches",
    ),
    0xA702: (
        STATUS_FAILURE,
        "Refused: Out of resources, unable to perform sub-operations",
    ),
    0xA801: (STATUS_FAILURE, "Move destination unknown"),
    0xA900: (STATUS_FAILURE, "Identifier does not match SOP class"),
    0xAA00: (
        STATUS_FAILURE,
        "None of the frames requested were found in the SOP instance",
    ),
    0xAA01: (STATUS_FAILURE, "Unable to create new object for this SOP class"),
    0xAA02: (STATUS_FAILURE, "Unable to extract frames"),
    0xAA03: (
        STATUS_FAILURE,
        "Time-based request received for a non-time-based original SOP Instance",
    ),
    0xAA04: (STATUS_FAILURE, "Invalid request"),
    0xFF00: (STATUS_PENDING, "Sub-operations are continuing"),
    0xB000: (STATUS_WARNING, "Sub-operations completed, one or more failures"),
}

# Ranged values
for _code in range(0xC000, 0xCFFF + 1):
    QR_MOVE_SERVICE_CLASS_STATUS[_code] = (STATUS_FAILURE, "Unable to Process")

# Add the General status code values - PS3.7 Annex C
QR_MOVE_SERVICE_CLASS_STATUS.update(GENERAL_STATUS)


# Query Retrieve - GET specific status code values
#   PS3.4 Annex C.4.3.1.4 and Annex Y.4.2.1.4
# Hanging Protocol  - GET specific status code values
#   PS3.4 Annex U.4.3
# Color Palette - GET specific status code values
#   PS3.4 Annex X
# Implant Template - GET specific status code values
#   PS3.4 Annex BB
# Defined Procedure Protocol - GET specific status code values
#   PS3.4 Annex HH
QR_GET_SERVICE_CLASS_STATUS: StatusDictType = {
    0xA701: (
        STATUS_FAILURE,
        "Refused: Out of resources, unable to calculate number of matches",
    ),
    0xA702: (
        STATUS_FAILURE,
        "Refused: Out of resources, unable to perform sub-operations",
    ),
    0xA900: (STATUS_FAILURE, "Identifier does not match SOP class"),
    0xAA00: (
        STATUS_FAILURE,
        "None of the frames requested were found in the SOP instance",
    ),
    0xAA01: (STATUS_FAILURE, "Unable to create new object for this SOP class"),
    0xAA02: (STATUS_FAILURE, "Unable to extract frames"),
    0xAA03: (
        STATUS_FAILURE,
        "Time-based request received for a non-time-based original SOP Instance",
    ),
    0xAA04: (STATUS_FAILURE, "Invalid request"),
    0xFF00: (STATUS_PENDING, "Sub-operations are continuing"),
    0xB000: (
        STATUS_WARNING,
        "Sub-operations complete, one or more failures or warnings",
    ),
}

# Ranged values
for _code in range(0xC000, 0xCFFF + 1):
    QR_GET_SERVICE_CLASS_STATUS[_code] = (STATUS_FAILURE, "Unable to Process")

# Add the General status code values - PS3.7 Annex C
QR_GET_SERVICE_CLASS_STATUS.update(GENERAL_STATUS)


# Modality Worklist Service Class specific status code values
#   PS3.4 Annex K
MODALITY_WORKLIST_SERVICE_CLASS_STATUS: StatusDictType = {
    0xA700: (STATUS_FAILURE, "Refused: Out of resources"),
    0xA900: (STATUS_FAILURE, "Identifier does not match SOP class"),
    0xFF00: (
        STATUS_PENDING,
        "Matches are continuing - current match is supplied and any "
        "Optional Keys were supported in the same manner as Required "
        "Keys",
    ),
    0xFF01: (
        STATUS_PENDING,
        "Matches are continuing - warning that one or more Optional "
        "Keys were not supported for existence for this Identifier",
    ),
}

# Ranged values
for _code in range(0xC000, 0xCFFF + 1):
    MODALITY_WORKLIST_SERVICE_CLASS_STATUS[_code] = (
        STATUS_FAILURE,
        "Unable to Process",
    )

# Add the General status code values - PS3.7 Annex C
MODALITY_WORKLIST_SERVICE_CLASS_STATUS.update(GENERAL_STATUS)


# Relevant Patient Information Query Service Class specific status code values
# PS3.4 Annex Q
RELEVANT_PATIENT_SERVICE_CLASS_STATUS: StatusDictType = {
    0xA700: (STATUS_FAILURE, "Out of resources"),
    0xA900: (STATUS_FAILURE, "Identifier doesn't match SOP Class"),
    0xC000: (STATUS_FAILURE, "Unable to process"),
    0xC100: (STATUS_FAILURE, "More than one match found"),
    0xC200: (STATUS_FAILURE, "Unable to support requested template"),
    0xFF00: (STATUS_PENDING, "Current match is supplied"),
}
RELEVANT_PATIENT_SERVICE_CLASS_STATUS.update(GENERAL_STATUS)


# Substance Administration Query Service Class specific status code values
#   PS3.4 Annex V
SUBSTANCE_ADMINISTRATION_SERVICE_CLASS_STATUS: StatusDictType = {
    0xA700: (STATUS_FAILURE, "Out of resources"),
    0xA900: (STATUS_FAILURE, "Data set doesn't match SOP Class"),
    0xFF00: (
        STATUS_PENDING,
        "Matches are continuing, current match is supplied and any "
        "Optional Keys were supported in the same manner as Required "
        "Keys",
    ),
    0xFF01: (
        STATUS_PENDING,
        "Matches are continuing, warning that one or more Optional "
        "Keys were not supported for existence for this Identifier",
    ),
}

# Ranged values
for _code in range(0xC000, 0xCFFF + 1):
    SUBSTANCE_ADMINISTRATION_SERVICE_CLASS_STATUS[_code] = (
        STATUS_FAILURE,
        "Unable to Process",
    )

SUBSTANCE_ADMINISTRATION_SERVICE_CLASS_STATUS.update(GENERAL_STATUS)


# Non-Patient Object Storage Service Class specific status code values
#   PS3.4 Annex GG
NON_PATIENT_SERVICE_CLASS_STATUS: StatusDictType = {
    0xA700: (STATUS_FAILURE, "Out of resources"),
    0xA900: (STATUS_FAILURE, "Data set doesn't match SOP Class"),
    0xC000: (STATUS_FAILURE, "Cannot understand"),
}
NON_PATIENT_SERVICE_CLASS_STATUS.update(GENERAL_STATUS)


# Procedure Step SOP Class specific status code values
PROCEDURE_STEP_STATUS: StatusDictType = {
    0x0001: (STATUS_WARNING, "Requested optional attributes are not supported"),
    0x0110: (
        STATUS_FAILURE,
        "Performed Procedure Step object may no longer be updated",
    ),
}
PROCEDURE_STEP_STATUS.update(GENERAL_STATUS)


# Print Job Management Service Class specific status code values
PRINT_JOB_MANAGEMENT_SERVICE_CLASS_STATUS: StatusDictType = {
    0xB600: (STATUS_WARNING, "Memory allocation not supported"),
    0xB601: (STATUS_WARNING, "Film session printing (collation) is not supported"),
    0xB602: (
        STATUS_WARNING,
        "Film Session SOP Instance hierarchy does not contain Image "
        "Box SOP Instances (empty page)",
    ),
    0xB603: (
        STATUS_WARNING,
        "Film Box SOP Instance hierarchy does not contain Image Box "
        "SOP Instances (empty page)",
    ),
    0xB604: (
        STATUS_WARNING,
        "Image size is larger than image box size, the image has been demagnified",
    ),
    0xB605: (
        STATUS_WARNING,
        "Requested minimum density or maximum density outside of "
        "printer's operating range. The print will use its respective "
        "minimum or maximum density value instead",
    ),
    0xB609: (
        STATUS_WARNING,
        "Image size is larger than the image box size, the image has "
        "been cropped to fit",
    ),
    0xB60A: (
        STATUS_WARNING,
        "Image size or Combined Print Image size is larger than the "
        "image box size, image or combined print image has been "
        "decimated to fit",
    ),
    0xC600: (
        STATUS_FAILURE,
        "Film Session SOP Instance hierarchy does not contain Film Box "
        "SOP Instances",
    ),
    0xC601: (
        STATUS_FAILURE,
        "Unable to create Print Job SOP Instance; print queue is full",
    ),
    0xC602: (
        STATUS_FAILURE,
        "Unable to create Print Job SOP instance; print queue is full",
    ),
    0xC603: (STATUS_FAILURE, "Image size is larger than image box size"),
    0xC605: (STATUS_FAILURE, "Insufficient memory in printer to store the image"),
    0xC613: (
        STATUS_FAILURE,
        "Combined print image size is larger than the image box size",
    ),
    0xC616: (
        STATUS_FAILURE,
        "There is an existing film box that has not been printed and "
        "N-ACTION at the film session level is not supported. A new "
        "film box will not be created when a previous film box has not "
        "been printed",
    ),
}
PRINT_JOB_MANAGEMENT_SERVICE_CLASS_STATUS.update(GENERAL_STATUS)


# Storage Commitment Service Class specific status code values
STORAGE_COMMITMENT_SERVICE_CLASS_STATUS = GENERAL_STATUS


# Storage Management Service Class specific status code values
STORAGE_MANAGEMENT_SERVICE_CLASS_STATUS: StatusDictType = {
    0xB010: (
        STATUS_WARNING,
        (
            "Attribute list error - One or more of Key Attributes are not "
            "supported for matching"
        ),
    ),
}
STORAGE_MANAGEMENT_SERVICE_CLASS_STATUS.update(GENERAL_STATUS)


# Application Event Logging Service Class specific status code values
APPLICATION_EVENT_LOGGING_SERVICE_CLASS_STATUS: StatusDictType = {
    0xB101: (
        STATUS_WARNING,
        "Specified Synchronisation Frame of Reference UID doesn't "
        "match SCP Synchronization Frame of Reference",
    ),
    0xB102: (
        STATUS_WARNING,
        "Study Instance UID coercion; event logged under a different "
        "Study Instance UID",
    ),
    0xB104: (
        STATUS_WARNING,
        "IDs inconsistent in matching a current study; event logged",
    ),
    0xC101: (
        STATUS_FAILURE,
        "Procedural logging not available for specified Study Instance UID",
    ),
    0xC102: (STATUS_FAILURE, "Event Information doesn't match template"),
    0xC103: (STATUS_FAILURE, "Cannot match event to a current study"),
    0xC104: (
        STATUS_FAILURE,
        "IDs inconsistent in matching a current study; event not logged",
    ),
    0xC10E: (
        STATUS_FAILURE,
        "Operator not authorised to add entry to Medication Administration Record",
    ),
    0xC110: (
        STATUS_FAILURE,
        "Patient cannot be identified from Patient ID (0010,0020) or "
        "Admission ID (0038,0010)",
    ),
    0xC111: (STATUS_FAILURE, "Update of Medication Administration Record failed"),
}
APPLICATION_EVENT_LOGGING_SERVICE_CLASS_STATUS.update(GENERAL_STATUS)


# Media Creation Management Service Class specific status code values
MEDIA_CREATION_MANAGEMENT_SERVICE_CLASS_STATUS: StatusDictType = {
    0x0001: (STATUS_WARNING, "Requested optional Attributes are not supported"),
    0xA510: (
        STATUS_FAILURE,
        "An Initiate Media Creation action has already been received "
        "for this SOP Instance",
    ),
    0xC201: (STATUS_FAILURE, "Media creation request already completed"),
    0xC202: (
        STATUS_FAILURE,
        "Media creation request already in progress and cannot be interrupted",
    ),
    0xC203: (STATUS_FAILURE, "Cancellation denied for unspecified reason"),
}
MEDIA_CREATION_MANAGEMENT_SERVICE_CLASS_STATUS.update(GENERAL_STATUS)


# Unified Procedure Step Service specific status code values
UNIFIED_PROCEDURE_STEP_SERVICE_CLASS_STATUS: StatusDictType = {}
# Ranged values
for _code in range(0xC000, 0xCFFF + 1):
    UNIFIED_PROCEDURE_STEP_SERVICE_CLASS_STATUS[_code] = (
        STATUS_FAILURE,
        "Unable to Process",
    )

UNIFIED_PROCEDURE_STEP_SERVICE_CLASS_STATUS.update(
    {
        0x0001: (STATUS_WARNING, "Requested optional attributes are not supported"),
        0xA700: (STATUS_FAILURE, "Out of resources"),
        0xA900: (STATUS_FAILURE, "Identifier doesn't match SOP Class"),
        0xB300: (STATUS_WARNING, "The UPS was created with modifications"),
        0xB301: (STATUS_WARNING, "Deletion lock not granted"),
        0xB304: (
            STATUS_WARNING,
            "The UPS is already in the requested state of CANCELED",
        ),
        0xB305: (STATUS_WARNING, "Coerced invalid values to valid values"),
        0xB306: (
            STATUS_WARNING,
            "The UPS is already in the requested state of COMPLETED",
        ),
        0xC300: (STATUS_FAILURE, "The UPS may no longer be updated"),
        0xC301: (STATUS_FAILURE, "The correct Transaction UID was not provided"),
        0xC302: (STATUS_FAILURE, "The UPS is already IN PROGRESS"),
        0xC303: (
            STATUS_FAILURE,
            "The UPS may only become SCHEDULED via N-CREATE, not N-SET or N-ACTION",
        ),
        0xC304: (
            STATUS_FAILURE,
            "The UPS has not met final state requirements for the "
            "requested state change",
        ),
        0xC307: (
            STATUS_FAILURE,
            "Specified SOP Instance UID does not exist or is not a UPS "
            "Instance managed by this SCP",
        ),
        0xC308: (STATUS_FAILURE, "Receiving AE-TITLE is unknown to this SCP"),
        0xC309: (STATUS_FAILURE, "The provided value of UPS State was not SCHEDULED"),
        0xC310: (STATUS_FAILURE, "The UPS is not yet in the IN PROGRESS state"),
        0xC311: (STATUS_FAILURE, "The UPS is already COMPLETED"),
        0xC312: (STATUS_FAILURE, "The performer cannot be contacted"),
        0xC313: (STATUS_FAILURE, "Performer chooses not to cancel"),
        0xC314: (
            STATUS_FAILURE,
            "Specified action is not appropriate for specified instance",
        ),
        0xC315: (STATUS_FAILURE, "SCP does not support Event Reports"),
        0xFF00: (
            STATUS_PENDING,
            "Matches are continuing - current match is supplied an any "
            "Optional Keys were supported in the same manner as Required "
            "Keys",
        ),
        0xFF01: (
            STATUS_PENDING,
            "Matches are continuing - current match is supplied an any "
            "Optional Keys were not supported for existence for this "
            "Identifier",
        ),
    }
)
UNIFIED_PROCEDURE_STEP_SERVICE_CLASS_STATUS.update(GENERAL_STATUS)


# RT Machine Verification Service Class specific status code values
RT_MACHINE_VERIFICATION_SERVICE_CLASS_STATUS: StatusDictType = {
    0xC112: (
        STATUS_FAILURE,
        "No such object instance - applicable Machine Verification "
        "instance not found",
    ),
    0xC221: (
        STATUS_FAILURE,
        "The Referenced Fraction Group Number does not exist in the referenced plan",
    ),
    0xC222: (STATUS_FAILURE, "No beams exist within the referenced fraction group"),
    0xC223: (
        STATUS_FAILURE,
        "SCU already verifying and cannot currently process this request",
    ),
    0xC224: (
        STATUS_FAILURE,
        "Referenced Beam Number not found within the referenced Fraction Group",
    ),
    0xC225: (STATUS_FAILURE, "Referenced device or accessory not supported"),
    0xC226: (
        STATUS_FAILURE,
        "Referenced device or accessory not found within the referenced beam",
    ),
    0xC227: (STATUS_FAILURE, "No such object instance - Referenced RT Plan not found"),
}
RT_MACHINE_VERIFICATION_SERVICE_CLASS_STATUS.update(GENERAL_STATUS)


def code_to_status(code: int) -> Dataset:
    """Return a :class:`~pydicom.dataset.Dataset` with a *Status* element
    value of `code`.
    """
    if isinstance(code, int) and code >= 0:
        ds = Dataset()
        ds.Status = code
        return ds

    raise ValueError("'code' must be a positive integer.")


def code_to_category(code: int) -> str:
    """Return a *Status* category as :class:`str` or ``'Unknown'`` if not
    recognised.
    """
    # pylint: disable=too-many-return-statements
    if isinstance(code, int) and code >= 0:
        if code == 0x0000:
            return STATUS_SUCCESS

        if code in [0xFF00, 0xFF01]:
            return STATUS_PENDING

        if code == 0xFE00:
            return STATUS_CANCEL

        if code in [
            0x0105,
            0x0106,
            0x0110,
            0x0111,
            0x0112,
            0x0113,
            0x0114,
            0x0115,
            0x0117,
            0x0118,
            0x0119,
            0x0120,
            0x0121,
            0x0122,
            0x0123,
            0x0124,
            0x0210,
            0x0211,
            0x0212,
            0x0213,
        ]:
            return STATUS_FAILURE

        if code in range(0xA000, 0xB000):
            return STATUS_FAILURE

        if code in range(0xC000, 0xD000):
            return STATUS_FAILURE

        if code in [0x0107, 0x0116]:
            return STATUS_WARNING

        if code in range(0xB000, 0xC000):
            return STATUS_WARNING

        if code == 0x0001:
            return STATUS_WARNING

        return STATUS_UNKNOWN

    raise ValueError("'code' must be a positive integer.")


class Status(IntEnum):
    """Constants for common status codes.

    .. versionadded:: 1.5

    New constants can be added with the ``Status.add(name, code)`` method but
    the documentation for it is missing due to a bug in Sphinx. `name` is
    the variable name of the constant to add as a :class:`str` and `code` is
    the corresponding status code as an :class:`int`.

    Examples
    --------

    ::

        from pynetdicom.status import Status

        # Customise the class
        Status.add('UNABLE_TO_PROCESS', 0xC000)

        def handle_store(event):
            try:
                event.dataset.save_as('temp.dcm')
            except:
                return Status.UNABLE_TO_PROCESS

            return Status.SUCCESS

    """

    SUCCESS = 0x0000
    """``0x0000`` - Success"""
    CANCEL = 0xFE00
    """``0xFE00`` - Operation terminated"""
    PENDING = 0xFF00
    """``0xFF00`` - Matches or sub-operations are continuing"""
    MOVE_DESTINATION_UNKNOWN = 0xA801
    """``0xA801`` - Move destination unknown"""

    @classmethod
    def add(cls, name: str, code: int) -> None:
        """Add a new constant to `Status`.

        Parameters
        ----------
        name : str
            The name of the constant to add.
        code : int
            The status code corresponding to the name.
        """
        setattr(cls, name, code)
